from __future__ import annotations

import inspect
import urllib.parse
from collections import OrderedDict
from typing import TYPE_CHECKING, Any, Callable

import clearskies.column
import clearskies.configs
import clearskies.configurable
import clearskies.di
import clearskies.end
import clearskies.parameters_to_properties
import clearskies.typing
from clearskies import autodoc, exceptions
from clearskies.authentication import Authentication, Authorization, Public
from clearskies.autodoc import schema
from clearskies.autodoc.request import Parameter, Request
from clearskies.autodoc.response import Response
from clearskies.functional import routing, string, validations

if TYPE_CHECKING:
    from clearskies import Column, Model, SecurityHeader
    from clearskies.input_outputs import InputOutput
    from clearskies.schema import Schema
    from clearskies.security_headers import Cors


class Endpoint(
    clearskies.end.End,  # type: ignore
    clearskies.configurable.Configurable,
    clearskies.di.InjectableProperties,
):
    """
    Endpoints - the clearskies workhorse.

    With clearskies, endpoints exist to offload some drudgery and make your life easier, but they can also
    get out of your way when you don't need them.  Think of them as pre-built endpoints that can execute
    common functionality needed for web applications/APIs.  Instead of defining a function that fetches
    records from your backend and returns them to the end user, you can let the list endpoint do this for you
    with a minimal amount of configuration.  Instead of making an endpoint that creates records, just deploy
    a create endpoint.  Each endpoint has their own configuration settings, but there are some configuration
    settings that are common to all endpoints, which are listed below:
    """

    """
    The dependency injection container
    """
    di = clearskies.di.inject.Di()

    """
    Whether or not this endpoint can handle CORS
    """
    has_cors = False

    """
    The actual CORS header
    """
    cors_header: Cors | None = None

    """
    Set some response headers that should be returned for this endpoint.

    Provide a list of response headers to return to the caller when this endpoint is executed.
    This should be given a list containing a combination of strings or callables that return a list of strings.
    The strings in question should be headers formatted as "key: value".  If you attach a callable, it can accept
    any of the standard dependencies or context-specific values like any other callable in a clearskies
    application:

    ```python
    def custom_headers(query_parameters):
        some_value = "yes" if query_parameters.get("stuff") else "no"
        return [f"x-custom: {some_value}", "content-type: application/custom"]

    endpoint = clearskies.endpoints.Callable(
        lambda: {"hello": "world"},
        response_headers=custom_headers,
    )

    wsgi = clearskies.contexts.WsgiRef(endpoint)
    wsgi()
    ```
    """
    response_headers = clearskies.configs.StringListOrCallable(default=[])

    """
    Set the URL for the endpoint

    When an endpoint is attached directly to a context, then the endpoint's URL becomes the exact URL
    to invoke the endpoint.  If it is instead attached to an endpoint group, then the URL of the endpoint
    becomes a suffix on the URL of the group.  This is described in more detail in the documentation for endpoint
    groups, so here's an example of attaching endpoints directly and setting the URL:

    ```python
    import clearskies

    endpoint = clearskies.endpoints.Callable(
        lambda: {"hello": "World"},
        url="/hello/world",
    )

    wsgi = clearskies.contexts.WsgiRef(endpoint)
    wsgi()
    ```

    Which then acts as expected:

    ```bash
    $ curl 'http://localhost:8080/hello/asdf' | jq
    {
        "status": "client_error",
        "error": "Not Found",
        "data": [],
        "pagination": {},
        "input_errors": {}
    }

    $ curl 'http://localhost:8080/hello/world' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "hello": "world"
        },
        "pagination": {},
        "input_errors": {}
    }
    ```

    Some endpoints allow or require the use of named routing parameters.  Named routing paths are created using either the
    `/{name}/` syntax or `/:name/`.  These parameters can be injected into any callable via the `routing_data`
    dependency injection name, as well as via their name:

    ```python
    import clearskies

    endpoint = clearskies.endpoints.Callable(
        lambda first_name, last_name: {"hello": f"{first_name} {last_name}"},
        url="/hello/:first_name/{last_name}",
    )

    wsgi = clearskies.contexts.WsgiRef(endpoint)
    wsgi()
    ```

    Which you can then invoke in the usual way:

    ```bash
    $ curl 'http://localhost:8080/hello/bob/brown' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "hello": "bob brown"
        },
        "pagination": {},
        "input_errors": {}
    }

    ```

    """
    url = clearskies.configs.Url(default="")

    """
    The allowed request methods for this endpoint.

    By default, only GET is allowed.

    ```python
    import clearskies

    endpoint = clearskies.endpoints.Callable(
        lambda: {"hello": "world"},
        request_methods=["POST"],
    )

    wsgi = clearskies.contexts.WsgiRef(endpoint)
    wsgi()
    ```

    And to execute:

    ```bash
    $ curl 'http://localhost:8080/' -X POST | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "hello": "world"
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl 'http://localhost:8080/' -X GET | jq
    {
        "status": "client_error",
        "error": "Not Found",
        "data": [],
        "pagination": {},
        "input_errors": {}
    }
    ```
    """
    request_methods = clearskies.configs.SelectList(
        allowed_values=["GET", "POST", "PUT", "DELETE", "PATCH", "QUERY"], default=["GET"]
    )

    """
    The authentication for this endpoint (default is public)

    Use this to attach an instance of `clearskies.authentication.Authentication` to an endpoint, which enforces authentication.
    For more details, see the dedicated documentation section on authentication and authorization. By default, all endpoints are public.
    """
    authentication = clearskies.configs.Authentication(default=Public())

    """
    The authorization rules for this endpoint

    Use this to attach an instance of `clearskies.authentication.Authorization` to an endpoint, which enforces authorization.
    For more details, see the dedicated documentation section on authentication and authorization. By default, no authorization is enforced.
    """
    authorization = clearskies.configs.Authorization(default=Authorization())

    """
    An override of the default model-to-json mapping for endpoints that auto-convert models to json.

    Many endpoints allow you to return a model which is then automatically converted into a JSON response.  When this is the case,
    you can provide a callable in the `output_map` parameter which will be called instead of following the usual method for
    JSON conversion.  Note that if you use this method, you should also specify `output_schema`, which the autodocumentation
    will then use to document the endpoint.

    Your function can request any named dependency injection parameter as well as the standard context parameters for the request.

    ```python
    import clearskies
    import datetime
    from dateutil.relativedelta import relativedelta

    class User(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()
        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        dob = clearskies.columns.Datetime()

    class UserResponse(clearskies.Schema):
        id = clearskies.columns.String()
        name = clearskies.columns.String()
        age = clearskies.columns.Integer()
        is_special = clearskies.columns.Boolean()

    def user_to_json(model: User, utcnow: datetime.datetime, special_person: str):
        return {
            "id": model.id,
            "name": model.name,
            "age": relativedelta(utcnow, model.dob).years,
            "is_special": model.name.lower() == special_person.lower(),
        }

    list_users = clearskies.endpoints.List(
        model_class=User,
        url="/{special_person}",
        output_map = user_to_json,
        output_schema = UserResponse,
        readable_column_names=["id", "name"],
        sortable_column_names=["id", "name", "dob"],
        default_sort_column_name="dob",
        default_sort_direction="DESC",
    )

    wsgi = clearskies.contexts.WsgiRef(
        list_users,
        classes=[User],
        bindings={
            "special_person": "jane",
            "memory_backend_default_data": [
                {
                    "model_class": User,
                    "records": [
                        {"id": "1-2-3-4", "name": "Bob", "dob": datetime.datetime(1990, 1, 1)},
                        {"id": "1-2-3-5", "name": "Jane", "dob": datetime.datetime(2020, 1, 1)},
                        {"id": "1-2-3-6", "name": "Greg", "dob": datetime.datetime(1980, 1, 1)},
                    ]
                },
            ]
        }
    )
    wsgi()
    ```

    Which gives:

    ```bash
    $ curl 'http://localhost:8080/jane' | jq
    {
        "status": "success",
        "error": "",
        "data": [
            {
                "id": "1-2-3-5",
                "name": "Jane",
                "age": 5,
                "is_special": true
            }
            {
                "id": "1-2-3-4",
                "name": "Bob",
                "age": 35,
                "is_special": false
            },
            {
                "id": "1-2-3-6",
                "name": "Greg",
                "age": 45,
                "is_special": false
            },
        ],
        "pagination": {
            "number_results": 3,
            "limit": 50,
            "next_page": {}
        },
        "input_errors": {}
    }

    ```

    """
    output_map = clearskies.configs.Callable(default=None)

    """
    A schema that describes the expected output to the client.

    This is used to build the auto-documentation.  See the documentation for clearskies.endpoint.output_map for examples.
    Note that this is typically not required - when returning models and relying on clearskies to auto-convert to JSON,
    it will also automatically generate your documentation.
    """
    output_schema = clearskies.configs.Schema(default=None)

    """
    The model class used by this endpoint.

    The majority of endpoints require a model class that tells the endpoint where to get/save its data.
    """
    model_class = clearskies.configs.ModelClass(default=None)

    """
    Columns from the model class that should be returned to the client.

    Most endpoints use a model to build the return response to the user.  In this case, `readable_column_names`
    instructs the model what columns should be sent back to the user.  This information is similarly used when generating
    the documentation for the endpoint.

    ```python
    import clearskies

    class User(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()
        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        secret = clearskies.columns.String()

    list_users = clearskies.endpoints.List(
        model_class=User,
        readable_column_names=["id", "name"],
        sortable_column_names=["id", "name"],
        default_sort_column_name="name",
    )

    wsgi = clearskies.contexts.WsgiRef(
        list_users,
        classes=[User],
        bindings={
            "memory_backend_default_data": [
                {
                    "model_class": User,
                    "records": [
                        {"id": "1-2-3-4", "name": "Bob", "secret": "Awesome dude"},
                        {"id": "1-2-3-5", "name": "Jane", "secret": "Gets things done"},
                        {"id": "1-2-3-6", "name": "Greg", "secret": "Loves chocolate"},
                    ]
                },
            ]
        }
    )
    wsgi()
    ```

    And then:

    ```bash
    $ curl 'http://localhost:8080'
    {
        "status": "success",
        "error": "",
        "data": [
            {
                "id": "1-2-3-4",
                "name": "Bob"
            },
            {
                "id": "1-2-3-6",
                "name": "Greg"
            },
            {
                "id": "1-2-3-5",
                "name": "Jane"
            }
        ],
        "pagination": {
            "number_results": 3,
            "limit": 50,
            "next_page": {}
        },
        "input_errors": {}
    }

    ```
    """
    readable_column_names = clearskies.configs.ReadableModelColumns("model_class", default=[])

    """
    Specifies which columns from a model class can be set by the client.

    Many endpoints allow or require input from the client.  The most common way to provide input validation
    is by setting the model class and using `writeable_column_names` to specify which columns the end client can
    set.  Clearskies will then use the model schema to validate the input and also auto-generate documentation
    for the endpoint.

    ```python
    import clearskies

    class User(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()
        id = clearskies.columns.Uuid()
        name = clearskies.columns.String(validators=[clearskies.validators.Required()])
        date_of_birth = clearskies.columns.Date()

    send_user = clearskies.endpoints.Callable(
        lambda request_data: request_data,
        request_methods=["GET","POST"],
        writeable_column_names=["name", "date_of_birth"],
        model_class=User,
    )

    wsgi = clearskies.contexts.WsgiRef(send_user)
    wsgi()
    ```

    If we send a valid payload:

    ```bash
    $ curl 'http://localhost:8080' -d '{"name":"Jane","date_of_birth":"01/01/1990"}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "name": "Jane",
            "date_of_birth": "01/01/1990"
        },
        "pagination": {},
        "input_errors": {}
    }
    ```

    And we can see the automatic input validation by sending some incorrect data:

    ```bash
    $ curl 'http://localhost:8080' -d '{"name":"","date_of_birth":"this is not a date","id":"hey"}' | jq
    {
        "status": "input_errors",
        "error": "",
        "data": [],
        "pagination": {},
        "input_errors": {
            "name": "'name' is required.",
            "date_of_birth": "given value did not appear to be a valid date",
            "other_column": "Input column other_column is not an allowed input column."
        }
    }
    ```

    """
    writeable_column_names = clearskies.configs.WriteableModelColumns("model_class", default=[])

    """
    Columns from the model class that can be searched by the client.

    Sets which columns the client is allowed to search (for endpoints that support searching).
    """
    searchable_column_names = clearskies.configs.SearchableModelColumns("model_class", default=[])

    """
    A function to call to add custom input validation logic.

    Typically, input validation happens by choosing the appropriate column in your schema and adding validators where necessary.  You
    can also create custom columns with their own input validation logic.  However, if desired, endpoints that accept user input also
    allow you to add callables for custom validation logic.  These functions should return a dictionary where the key name
    represents the name of the column that has invalid input, and the value is a human-readable error message.  If no input errors are
    found, then the callable should return an empty dictionary.  As usual, the callable can request any standard dependencies configured
    in the dependency injection container or proivded by input_output.get_context_for_callables.

    Note that most endpoints (such as Create and Update) explicitly require input.  As a result, if a request comes in without input
    from the end user, it will be rejected before calling your input validator.  In these cases you can depend on request_data always
    being a dictionary.  The Callable endpoint, however, only requires input if `writeable_column_names` is set.  If it's not set,
    and the end-user doesn't provide a request body, then request_data will be None.

    ```python
    import clearskies

    def check_input(request_data):
        if not request_data:
            return {}
        if request_data.get("name"):
            return {"name":"This is a privacy-preserving system, so please don't tell us your name"}
        return {}

    send_user = clearskies.endpoints.Callable(
        lambda request_data: request_data,
        request_methods=["GET", "POST"],
        input_validation_callable=check_input,
    )

    wsgi = clearskies.contexts.WsgiRef(send_user)
    wsgi()
    ```

    And when invoked:

    ```bash
    $ curl http://localhost:8080 -d '{"name":"sup"}' | jq
    {
        "status": "input_errors",
        "error": "",
        "data": [],
        "pagination": {},
        "input_errors": {
            "name": "This is a privacy-preserving system, so please don't tell us your name"
        }
    }

    $ curl http://localhost:8080 -d '{"hello":"world"}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "hello": "world"
        },
        "pagination": {},
        "input_errors": {}
    }
    ```

    """
    input_validation_callable = clearskies.configs.Callable(default=None)

    """
    A dictionary with columns that should override columns in the model.

    This is typically used to change column definitions on specific endpoints to adjust behavior: for intstance a model might use a `created_by_*`
    column to auto-populate some data, but an admin endpoint may need to override that behavior so the user can set it directly.

    This should be a dictionary with the column name as a key and the column itself as the value.  Note that you cannot use this to remove
    columns from the model.  In general, if you want a column not to be exposed through an endpoint, then all you have to do is remove
    that column from the list of writeable columns.

    ```python
    import clearskies

    endpoint = clearskies.Endpoint(
        column_overrides = {
            "name": clearskies.columns.String(validators=clearskies.validators.Required()),
        }
    )
    ```
    """
    column_overrides = clearskies.configs.Columns(default={})

    """
    Used in conjunction with external_casing to change the casing of the key names in the outputted JSON of the endpoint.

    To use these, set internal_casing to the casing scheme used in your model, and then set external_casing to the casing
    scheme you want for your API endpoints.  clearskies will then automatically convert all output key names accordingly.
    Note that for callables, this only works when you return a model and set `readable_columns`.  If you set `writeable_columns`,
    it will also map the incoming data.

    The allowed casing schemas are:

     1. `snake_case`
     2. `camelCase`
     3. `TitleCase`

    By default internal_casing and external_casing are both set to 'snake_case', which means that no conversion happens.

    ```python
    import clearskies
    import datetime

    class User(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()
        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        date_of_birth = clearskies.columns.Date()

    send_user = clearskies.endpoints.Callable(
        lambda users: users.create({"name":"Example","date_of_birth": datetime.datetime(2050, 1, 15)}),
        readable_column_names=["name", "date_of_birth"],
        internal_casing="snake_case",
        external_casing="TitleCase",
        model_class=User,
    )

    # because we're using name-based injection in our lambda callable (instead of type hinting) we have to explicitly
    # add the user model to the dependency injection container
    wsgi = clearskies.contexts.WsgiRef(send_user, classes=[User])
    wsgi()
    ```

    And then when called:

    ```bash
    $ curl http://localhost:8080  | jq
    {
        "Status": "Success",
        "Error": "",
        "Data": {
            "Name": "Example",
            "DateOfBirth": "2050-01-15"
        },
        "Pagination": {},
        "InputErrors": {}
    }
    ```
    """
    internal_casing = clearskies.configs.Select(["snake_case", "camelCase", "TitleCase"], default="snake_case")

    """
    Used in conjunction with internal_casing to change the casing of the key names in the outputted JSON of the endpoint.

    See the docs for `internal_casing` for more details and usage examples.
    """
    external_casing = clearskies.configs.Select(["snake_case", "camelCase", "TitleCase"], default="snake_case")

    """
    Configure standard security headers to be sent along in the response from this endpoint.

    Note that, with CORS, you generally only have to specify the origin.  The routing system will automatically add
    in the appropriate HTTP verbs, and the authorization classes will add in the appropriate headers.

    ```python
    import clearskies

    hello_world = clearskies.endpoints.Callable(
        lambda: {"hello": "world"},
        request_methods=["PATCH", "POST"],
        authentication=clearskies.authentication.SecretBearer(environment_key="MY_SECRET"),
        security_headers=[
            clearskies.security_headers.Hsts(),
            clearskies.security_headers.Cors(origin="https://example.com"),
        ],
    )

    wsgi = clearskies.contexts.WsgiRef(hello_world)
    wsgi()
    ```

    And then execute the options endpoint to see all the security headers:

    ```bash
    $ curl -v http://localhost:8080 -X OPTIONS
    * Host localhost:8080 was resolved.
    < HTTP/1.0 200 Ok
    < Server: WSGIServer/0.2 CPython/3.11.6
    < ACCESS-CONTROL-ALLOW-METHODS: PATCH, POST
    < ACCESS-CONTROL-ALLOW-HEADERS: Authorization
    < ACCESS-CONTROL-MAX-AGE: 5
    < ACCESS-CONTROL-ALLOW-ORIGIN: https://example.com
    < STRICT-TRANSPORT-SECURITY: max-age=31536000 ;
    < CONTENT-TYPE: application/json; charset=UTF-8
    < Content-Length: 0
    <
    * Closing connection
    ```

    """
    security_headers = clearskies.configs.SecurityHeaders(default=[])

    """
    A description for this endpoint.  This is added to any auto-documentation
    """
    description = clearskies.configs.String(default="")

    """
    Whether or not the routing data should also be persisted to the model.  Defaults to False.

    Note: this is only relevant for handlers that accept request data
    """
    include_routing_data_in_request_data = clearskies.configs.Boolean(default=False)

    """
    Additional conditions to always add to the results.

    where should be a single item or a list of items containing one of three things:

      1. Conditions expressed as a string (e.g. `"name=example"`, `"age>5"`)
      2. Queries built with a column (e.g. `SomeModel.name.equals("example")`, `SomeModel.age.greater_than(5)`)
      3. A callable which accepts and returns the mode (e.g. `lambda model: model.where("name=example")`)

    Here's an example:

    ```python
    import clearskies

    class Student(clearskies.Model):
        backend = clearskies.backends.MemoryBackend()
        id_column_name = "id"

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        grade = clearskies.columns.Integer()
        will_graduate = clearskies.columns.Boolean()

    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.List(
            Student,
            readable_column_names=["id", "name", "grade"],
            sortable_column_names=["name", "grade"],
            default_sort_column_name="name",
            where=["grade<10", Student.will_graduate.equals(True)],
        ),
        bindings={
            "memory_backend_default_data": [
                {
                    "model_class": Student,
                    "records": [
                        {"id": "1-2-3-4", "name": "Bob", "grade": 5, "will_graduate": True},
                        {"id": "1-2-3-5", "name": "Jane", "grade": 3, "will_graduate": True},
                        {"id": "1-2-3-6", "name": "Greg", "grade": 3, "will_graduate": False},
                        {"id": "1-2-3-7", "name": "Bob", "grade": 2, "will_graduate": True},
                        {"id": "1-2-3-8", "name": "Ann", "grade": 12, "will_graduate": True},
                    ],
                },
            ],
        },
    )
    wsgi()
    ```

    Which you can invoke:

    ```bash
    $ curl 'http://localhost:8080/' | jq
    {
        "status": "success",
        "error": "",
        "data": [
            {
                "id": "1-2-3-4",
                "name": "Bob",
                "grade": 5
            },
            {
                "id": "1-2-3-7",
                "name": "Bob",
                "grade": 2
            },
            {
                "id": "1-2-3-5",
                "name": "Jane",
                "grade": 3
            }
        ],
        "pagination": {},
        "input_errors": {}
    }
    ```
    and note that neither Greg nor Ann are returned.  Ann because she doesn't make the grade criteria, and Greg because
    he won't graduate.
    """
    where = clearskies.configs.Conditions(default=[])

    """
    Additional joins to always add to the query.

    ```python
    import clearskies

    class Student(clearskies.Model):
        backend = clearskies.backends.MemoryBackend()
        id_column_name = "id"

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        grade = clearskies.columns.Integer()
        will_graduate = clearskies.columns.Boolean()

    class PastRecord(clearskies.Model):
        backend = clearskies.backends.MemoryBackend()
        id_column_name = "id"

        id = clearskies.columns.Uuid()
        student_id = clearskies.columns.BelongsToId(Student)
        school_name = clearskies.columns.String()

    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.List(
            Student,
            readable_column_names=["id", "name", "grade"],
            sortable_column_names=["name", "grade"],
            default_sort_column_name="name",
            joins=["INNER JOIN past_records ON past_records.student_id=students.id"],
        ),
        bindings={
            "memory_backend_default_data": [
                {
                    "model_class": Student,
                    "records": [
                        {"id": "1-2-3-4", "name": "Bob", "grade": 5, "will_graduate": True},
                        {"id": "1-2-3-5", "name": "Jane", "grade": 3, "will_graduate": True},
                        {"id": "1-2-3-6", "name": "Greg", "grade": 3, "will_graduate": False},
                        {"id": "1-2-3-7", "name": "Bob", "grade": 2, "will_graduate": True},
                        {"id": "1-2-3-8", "name": "Ann", "grade": 12, "will_graduate": True},
                    ],
                },
                {
                    "model_class": PastRecord,
                    "records": [
                        {"id": "5-2-3-4", "student_id": "1-2-3-4", "school_name": "Best Academy"},
                        {"id": "5-2-3-5", "student_id": "1-2-3-5", "school_name": "Awesome School"},
                    ],
                },
            ],
        },
    )
    wsgi()
    ```

    Which when invoked:

    ```bash
    $ curl 'http://localhost:8080/' | jq
    {
        "status": "success",
        "error": "",
        "data": [
            {
                "id": "1-2-3-4",
                "name": "Bob",
                "grade": 5
            },
            {
                "id": "1-2-3-5",
                "name": "Jane",
                "grade": 3
            }
        ],
        "pagination": {},
        "input_errors": {}
    }
    ```

    e.g., the inner join reomves all the students that don't have an entry in the PastRecord model.

    """
    joins = clearskies.configs.Joins(default=[])

    cors_header: Cors = None  # type: ignore
    _model: clearskies.model.Model = None  # type: ignore
    _columns: dict[str, clearskies.column.Column] = None  # type: ignore
    _readable_columns: dict[str, clearskies.column.Column] = None  # type: ignore
    _writeable_columns: dict[str, clearskies.column.Column] = None  # type: ignore
    _searchable_columns: dict[str, clearskies.column.Column] = None  # type: ignore
    _sortable_columns: dict[str, clearskies.column.Column] = None  # type: ignore
    _as_json_map: dict[str, clearskies.column.Column] = None  # type: ignore

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        url: str = "",
        request_methods: list[str] = ["GET"],
        response_headers: list[str | Callable[..., list[str]]] = [],
        output_map: Callable[..., dict[str, Any]] | None = None,
        column_overrides: dict[str, Column] = {},
        internal_casing: str = "snake_case",
        external_casing: str = "snake_case",
        security_headers: list[SecurityHeader] = [],
        description: str = "",
        authentication: Authentication = Public(),
        authorization: Authorization = Authorization(),
    ):
        self.finalize_and_validate_configuration()
        for security_header in self.security_headers:
            if not security_header.is_cors:
                continue
            self.cors_header = security_header  # type: ignore
            self.has_cors = True
            break

    @property
    def model(self) -> Model:
        if self._model is None:
            self._model = self.di.build(self.model_class)
        return self._model

    @property
    def columns(self) -> dict[str, Column]:
        if self._columns is None:
            self._columns = self.model.get_columns()
        return self._columns

    @property
    def readable_columns(self) -> dict[str, Column]:
        if self._readable_columns is None:
            self._readable_columns = {name: self.columns[name] for name in self.readable_column_names}
        return self._readable_columns

    @property
    def writeable_columns(self) -> dict[str, Column]:
        if self._writeable_columns is None:
            self._writeable_columns = {name: self.columns[name] for name in self.writeable_column_names}
        return self._writeable_columns

    @property
    def searchable_columns(self) -> dict[str, Column]:
        if self._searchable_columns is None:
            self._searchable_columns = {name: self._columns[name] for name in self.sortable_column_names}
        return self._searchable_columns

    @property
    def sortable_columns(self) -> dict[str, Column]:
        if self._sortable_columns is None:
            self._sortable_columns = {name: self._columns[name] for name in self.sortable_column_names}
        return self._sortable_columns

    def get_request_data(self, input_output: InputOutput, required=True) -> dict[str, Any]:
        if not input_output.request_data:
            if input_output.has_body():
                raise exceptions.ClientError("Request body was not valid JSON")
            raise exceptions.ClientError("Missing required JSON body")
        if not isinstance(input_output.request_data, dict):
            raise exceptions.ClientError("Request body was not a JSON dictionary.")

        return {
            **input_output.request_data,  # type: ignore
            **(input_output.routing_data if self.include_routing_data_in_request_data else {}),
        }

    def fetch_model_with_base_query(self, input_output: InputOutput) -> Model:
        model = self.model
        for join in self.joins:
            if callable(join):
                model = self.di.call_function(join, model=model, **input_output.get_context_for_callables())
            else:
                model = model.join(join)
        for where in self.where:
            if callable(where):
                model = self.di.call_function(where, model=model, **input_output.get_context_for_callables())
            else:
                model = model.where(where)
        model = model.where_for_request(
            model,
            input_output.routing_data,
            input_output.authorization_data,
            input_output,
            overrides=self.column_overrides,
        )
        return self.authorization.filter_model(model, input_output.authorization_data, input_output)

    def handle(self, input_output: InputOutput) -> Any:
        raise NotImplementedError()

    def matches_request(self, input_output: InputOutput, allow_partial=False) -> bool:
        """Whether or not we can handle an incoming request based on URL and request method."""
        # soo..... this excessively duplicates the logic in __call__, but I'm being lazy right now
        # and not fixing it.
        request_method = input_output.get_request_method().upper()
        if request_method == "OPTIONS":
            return True
        if request_method not in self.request_methods:
            return False
        expected_url = self.url.strip("/")
        incoming_url = input_output.get_full_path().strip("/")
        if not expected_url and not incoming_url:
            return True

        matches, routing_data = routing.match_route(expected_url, incoming_url, allow_partial=allow_partial)
        return matches

    def populate_routing_data(self, input_output: InputOutput) -> Any:
        # matches_request is only checked by the endpoint group, not by the context.  As a result, we need to check our
        # route.  However we always have to check our route anyway because the full routing data can only be figured
        # out at the endpoint level, so calling out to routing.mattch_route is unavoidable.
        request_method = input_output.get_request_method().upper()
        if request_method == "OPTIONS":
            return self.cors(input_output)
        if request_method not in self.request_methods:
            return self.error(input_output, "Not Found", 404)
        expected_url = self.url.strip("/")
        incoming_url = input_output.get_full_path().strip("/")
        if expected_url or incoming_url:
            matches, routing_data = routing.match_route(expected_url, incoming_url, allow_partial=False)
            if not matches:
                return self.error(input_output, "Not Found", 404)
            input_output.routing_data = routing_data

    def failure(self, input_output: InputOutput) -> Any:
        return self.respond_json(input_output, {"status": "failure"}, 500)

    def input_errors(self, input_output: InputOutput, errors: dict[str, str], status_code: int = 200) -> Any:
        """Return input errors to the client."""
        return self.respond_json(input_output, {"status": "input_errors", "input_errors": errors}, status_code)

    def error(self, input_output: InputOutput, message: str, status_code: int) -> Any:
        """Return a client-side error (e.g. 400)."""
        return self.respond_json(input_output, {"status": "client_error", "error": message}, status_code)

    def redirect(self, input_output: InputOutput, location: str, status_code: int) -> Any:
        """Return a redirect."""
        input_output.response_headers.add("content-type", "text/html")
        input_output.response_headers.add("location", location)
        return self.respond(
            '<meta http-equiv="refresh" content="0; url=' + urllib.parse.quote(location) + '">Redirecting', status_code
        )

    def success(
        self,
        input_output: InputOutput,
        data: dict[str, Any] | list[Any],
        number_results: int | None = None,
        limit: int | None = None,
        next_page: Any = None,
    ) -> Any:
        """Return a successful response."""
        response_data = {"status": "success", "data": data, "pagination": {}}

        if next_page or number_results:
            if number_results is not None:
                for value in [number_results, limit]:
                    if value is not None and type(value) != int:
                        raise ValueError("number_results and limit must all be integers")

            response_data["pagination"] = {
                "number_results": number_results,
                "limit": limit,
                "next_page": next_page,
            }

        return self.respond_json(input_output, response_data, 200)

    def model_as_json(self, model: clearskies.model.Model, input_output: InputOutput) -> dict[str, Any]:
        if self.output_map:
            return self.di.call_function(self.output_map, model=model, **input_output.get_context_for_callables())

        if self._as_json_map is None:
            self._as_json_map = self._build_as_json_map(model)

        json = OrderedDict()
        for output_name, column in self._as_json_map.items():
            column_data = column.to_json(model)
            if len(column_data) == 1:
                json[output_name] = list(column_data.values())[0]
            else:
                for key, value in column_data.items():
                    json[self.auto_case_column_name(key, True)] = value
        return json

    def _build_as_json_map(self, model: clearskies.model.Model) -> dict[str, clearskies.column.Column]:
        conversion_map = {}
        if not self.readable_column_names:
            raise ValueError(
                "I was asked to convert a model to JSON but I wasn't provided with `readable_column_names'"
            )
        for column in self.readable_columns.values():
            conversion_map[self.auto_case_column_name(column.name, True)] = column
        return conversion_map

    def validate_input_against_schema(
        self, request_data: dict[str, Any], input_output: InputOutput, schema: Schema | type[Schema]
    ) -> None:
        if not self.writeable_column_names:
            raise ValueError(
                f"I was asked to validate input against a schema, but no writeable columns are defined, so I can't :(  This is probably a bug in the endpoint class - {self.__class__.__name__}."
            )
        request_data = self.map_request_data_external_to_internal(request_data)
        self.find_input_errors(request_data, input_output, schema)

    def map_request_data_external_to_internal(self, request_data, required=True):
        # we have to map from internal names to external names, because case mapping
        # isn't always one-to-one, so we want to do it exactly the same way that the documentation
        # is built.
        key_map = {self.auto_case_column_name(key, True): key for key in self.writeable_column_names}

        # and make sure we don't drop any data along the way, because the input validation
        # needs to return an error for unexpected data.
        return {key_map.get(key, key): value for (key, value) in request_data.items()}

    def find_input_errors(
        self, request_data: dict[str, Any], input_output: InputOutput, schema: Schema | type[Schema]
    ) -> None:
        input_errors: dict[str, str] = {}
        columns = schema.get_columns()
        model = self.di.build(schema) if inspect.isclass(schema) else schema
        for column_name in self.writeable_column_names:
            column = columns[column_name]
            input_errors = {
                **input_errors,
                **column.input_errors(model, request_data),  # type: ignore
            }
        input_errors = {
            **input_errors,
            **self.find_input_errors_from_callable(request_data, input_output),
        }
        for extra_column_name in set(request_data.keys()) - set(self.writeable_column_names):
            external_column_name = self.auto_case_column_name(extra_column_name, False)
            input_errors[external_column_name] = f"Input column {external_column_name} is not an allowed input column."
        if input_errors:
            raise exceptions.InputErrors(input_errors)

    def find_input_errors_from_callable(
        self, request_data: dict[str, Any] | list[Any] | None, input_output: InputOutput
    ) -> dict[str, str]:
        if not self.input_validation_callable:
            return {}

        more_input_errors = self.di.call_function(
            self.input_validation_callable, **input_output.get_context_for_callables()
        )
        if not isinstance(more_input_errors, dict):
            raise ValueError("The input error callable did not return a dictionary as required")
        return more_input_errors

    def cors(self, input_output: InputOutput):
        cors_header = self.cors_header if self.cors_header else Cors()
        for method in self.request_methods:
            cors_header.add_method(method)
        if self.authentication:
            self.authentication.set_headers_for_cors(cors_header)
        cors_header.set_headers_for_input_output(input_output)
        for security_header in self.security_headers:
            if security_header.is_cors:
                continue
            security_header.set_headers_for_input_output(input_output)
        return input_output.respond("", 200)

    def documentation(self) -> list[Request]:
        return []

    def documentation_components(self) -> dict[str, Any]:
        return {
            "models": self.documentation_models(),
            "securitySchemes": self.documentation_security_schemes(),
        }

    def documentation_security_schemes(self) -> dict[str, Any]:
        if not self.authentication or not self.authentication.documentation_security_scheme_name():
            return {}

        return {
            self.authentication.documentation_security_scheme_name(): (
                self.authentication.documentation_security_scheme()
            ),
        }

    def documentation_models(self) -> dict[str, schema.Schema]:
        return {}

    def documentation_pagination_response(self, include_pagination=True) -> schema.Schema:
        if not include_pagination:
            return schema.Object(self.auto_case_internal_column_name("pagination"), [], value={})
        model = self.di.build(self.model_class)
        return schema.Object(
            self.auto_case_internal_column_name("pagination"),
            [
                schema.Integer(self.auto_case_internal_column_name("number_results"), example=10),
                schema.Integer(self.auto_case_internal_column_name("limit"), example=100),
                schema.Object(
                    self.auto_case_internal_column_name("next_page"),
                    model.documentation_pagination_next_page_response(self.auto_case_internal_column_name),
                    model.documentation_pagination_next_page_example(self.auto_case_internal_column_name),
                ),
            ],
        )

    def documentation_success_response(
        self, data_schema: schema.Object | schema.Array, description: str = "", include_pagination: bool = False
    ) -> Response:
        return Response(
            200,
            schema.Object(
                "body",
                [
                    schema.String(self.auto_case_internal_column_name("status"), value="success"),
                    data_schema,
                    self.documentation_pagination_response(include_pagination=include_pagination),
                    schema.String(self.auto_case_internal_column_name("error"), value=""),
                    schema.Object(self.auto_case_internal_column_name("input_errors"), [], value={}),
                ],
            ),
            description=description,
        )

    def documentation_generic_error_response(self, description="Invalid Call", status=400) -> Response:
        return Response(
            status,
            schema.Object(
                "body",
                [
                    schema.String(self.auto_case_internal_column_name("status"), value="error"),
                    schema.Object(self.auto_case_internal_column_name("data"), [], value={}),
                    self.documentation_pagination_response(include_pagination=False),
                    schema.String(self.auto_case_internal_column_name("error"), example="User readable error message"),
                    schema.Object(self.auto_case_internal_column_name("input_errors"), [], value={}),
                ],
            ),
            description=description,
        )

    def documentation_input_error_response(self, description="Invalid client-side input") -> Response:
        email_example = self.auto_case_internal_column_name("email")
        return Response(
            200,
            schema.Object(
                "body",
                [
                    schema.String(self.auto_case_internal_column_name("status"), value="input_errors"),
                    schema.Object(self.auto_case_internal_column_name("data"), [], value={}),
                    self.documentation_pagination_response(include_pagination=False),
                    schema.String(self.auto_case_internal_column_name("error"), value=""),
                    schema.Object(
                        self.auto_case_internal_column_name("input_errors"),
                        [schema.String("[COLUMN_NAME]", example="User friendly error message")],
                        example={email_example: f"{email_example} was not a valid email address"},
                    ),
                ],
            ),
            description=description,
        )

    def documentation_access_denied_response(self) -> Response:
        return self.documentation_generic_error_response(description="Access Denied", status=401)

    def documentation_unauthorized_response(self) -> Response:
        return self.documentation_generic_error_response(description="Unauthorized", status=403)

    def documentation_not_found(self) -> Response:
        return self.documentation_generic_error_response(description="Not Found", status=404)

    def documentation_request_security(self):
        authentication = self.authentication
        name = authentication.documentation_security_scheme_name()
        return [{name: []}] if name else []

    def documentation_data_schema(
        self, schema: type[Schema] | None = None, column_names: list[str] = []
    ) -> list[schema.Schema]:
        if schema is None:
            schema = self.model_class
        if column_names is None and self.readable_column_names:
            readable_column_names: list[str] = self.readable_column_names
        properties = []

        columns = schema.get_columns()
        for column_name in readable_column_names:
            column = columns[column_name]
            for doc in column.documentation():
                doc.name = self.auto_case_internal_column_name(doc.name)
                properties.append(doc)

        return properties

    def standard_json_request_parameters(
        self, schema: type[Schema] | None = None, column_names: list[str] = []
    ) -> list[Parameter]:
        if not column_names:
            if not self.writeable_column_names:
                return []
            column_names = self.writeable_column_names

        if not schema:
            if not self.model_class:
                return []
            schema = self.model_class

        model_name = string.camel_case_to_snake_case(schema.__name__)
        columns = schema.get_columns()
        return [
            autodoc.request.JSONBody(
                columns[column_name].documentation(name=self.auto_case_column_name(column_name, True)),
                description=f"Set '{column_name}' for the {model_name}",
                required=columns[column_name].is_required,
            )
            for column_name in column_names
        ]

    def standard_url_request_parameters(self) -> list[Parameter]:
        parameter_names = routing.extract_url_parameter_name_map(self.url.strip("/"))
        return [
            autodoc.request.URLPath(
                autodoc.schema.String(parameter_name),
                description=f"The {parameter_name}.",
                required=True,
            )
            for parameter_name in parameter_names.keys()
        ]



import clearskies.configs
import clearskies.di
import clearskies.parameters_to_properties
from clearskies import autodoc
from clearskies.authentication.authentication import Authentication


class SecretBearer(Authentication, clearskies.di.InjectableProperties):
    """
    Secret Bearer performs authentication by checking against a static API key stored in either environment variables or a secret manager.

    This can be used in two different ways:

     1. Attached to an endpoint to enforce authentication
     2. Attached to an API backend to specify how to authenticate to the API endpoint.

    ### Authenticating Endpoints.

    When attached to an endpoint this will enforce authentication.  Clients authenticate themselves by providing the secret value
    via the `authorization` header.  In the following example we configure the secret bearer class to get the secretfrom an
    environment variable which is set in the code itself.  Normally you wouldn't set environment variables like this,
    but it's done here to create a self-contained example that is easy to run:

    ```python
    import os
    import clearskies

    os.environ["MY_AUTH_SECRET"] = "SUPERSECRET"

    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Callable(
            lambda: {"hello": "world"},
            authentication=clearskies.authentication.SecretBearer(environment_key="MY_AUTH_SECRET"),
        )
    )
    wsgi()
    ```
    We can then call it with and without the authentication header:

    ```bash
    $ curl 'http://localhost:8080' -H 'Authorization: SUPERSECRET' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "hello": "world"
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl 'http://localhost:8080' | jq
    {
        "status": "client_error",
        "error": "Not Authenticated",
        "data": [],
        "pagination": {},
        "input_errors": {}
    }

    $ curl 'http://localhost:8080' -H 'Authorization: NOTTHESECRET' | jq
    {
        "status": "client_error",
        "error": "Not Authenticated",
        "data": [],
        "pagination": {},
        "input_errors": {}
    }
    ```

    ### Authenticating to APIs

    The secret bearer class can also be attached to an API Backend to provide authentication to remote APIs.  To
    demonstrate, here is an example server that expects a secret token in the authorization header:

    ```python
    import os
    import clearskies
    from clearskies import columns

    os.environ["MY_SECRET"] = "SUPERSECRET"


    class Widget(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = columns.Uuid()
        name = columns.String()
        category = columns.String()
        cost = columns.Float()
        created_at = columns.Created()
        updated_at = columns.Updated()


    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.RestfulApi(
            url="widgets",
            model_class=Widget,
            authentication=clearskies.authentication.SecretBearer(environment_key="MY_SECRET"),
            readable_column_names=["id", "name", "category", "cost", "created_at", "updated_at"],
            writeable_column_names=["name", "category", "cost"],
            sortable_column_names=["name", "category", "cost"],
            searchable_column_names=["id", "name", "category", "cost"],
            default_sort_column_name="name",
        )
    )
    wsgi()
    ```

    Then here is a client app (you can launch the above server and then run this in a new terminal) that
    similarly uses the secret bearer class to authenticate to the server:

    ```python
    import os
    import clearskies
    from clearskies import columns

    os.environ["MY_SECRET"] = "SUPERSECRET"


    class Widget(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.ApiBackend(
            base_url="http://localhost:8080",
            authentication=clearskies.authentication.SecretBearer(environment_key="MY_SECRET"),
        )

        id = columns.String()
        name = columns.String()
        category = columns.String()
        cost = columns.Float()
        created_at = columns.Datetime()
        updated_at = columns.Datetime()


    def api_demo(widgets: Widget) -> Widget:
        thinga = widgets.create({"name": "Thinga", "category": "Doohickey", "cost": 125})
        mabob = widgets.create({"name": "Mabob", "category": "Doohicky", "cost": 150})
        return widgets


    cli = clearskies.contexts.Cli(
        clearskies.endpoints.Callable(
            api_demo,
            model_class=Widget,
            return_records=True,
            readable_column_names=["id", "name", "category", "cost", "created_at", "updated_at"],
        ),
        classes=[Widget],
    )
    cli()
    ```

    The above app declares a model class that matches the output from our server/api.  Note that the id,
    created_at, and updated_at columns all changed types to their "plain" types.  This is very normal.  The API
    is the one that is responsible for assigning ids and setting created/updated timestamps, so from the
    perspective of our client, these are plain string/datetime fields.  If we used the UUID or created/updated
    columns, then when the client called the API it would try to set all of these columns.  Since they are not
    writeable columns, the API would return an input error.  If you launch the above server/API and then run
    the given client script, you'll see output like this:

    ```json
    {
        "status": "success",
        "error": "",
        "data": [
            {
                "id": "54eef01d-7c87-4959-b525-dcb9047d9692",
                "name": "Mabob",
                "category": "Doohicky",
                "cost": 150.0,
                "created_at": "2025-06-13T15:19:27+00:00",
                "updated_at": "2025-06-13T15:19:27+00:00",
            },
            {
                "id": "ed1421b8-88ad-49d2-a130-c34b4ac4dfcf",
                "name": "Thinga",
                "category": "Doohickey",
                "cost": 125.0,
                "created_at": "2025-06-13T15:19:27+00:00",
                "updated_at": "2025-06-13T15:19:27+00:00",
            },
        ],
        "pagination": {},
        "input_errors": {},
    }
    ```
    """

    is_public = False
    can_authorize = False

    environment = clearskies.di.inject.Environment()
    secrets = clearskies.di.inject.Secrets()

    """
    The path in our secret manager from which the secret should be fetched.

    Of course, to use `secret_key`, you must also provide a secret manager.  The below example uses the dependency
    injection system to create a faux secret manager to demonstrate how it works in general:

    ```python
    from types import SimpleNamespace
    import clearskies

    def fetch_secret(path):
        if path == "/path/to/my/secret":
            return "SUPERSECRET"
        raise KeyError(f"Attempt to fetch non-existent secret: {path}")

    fake_secret_manager = SimpleNamespace(get=fetch_secret)

    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Callable(
            lambda: {"hello": "world"},
            authentication=clearskies.authentication.SecretBearer(secret_key="/path/to/my/secret"),
        ),
        bindings={
            "secrets": fake_secret_manager,
        },
    )
    wsgi()
    ```

    And when invoked:

    ```bash
    $ curl 'http://localhost:8080/' -H "Authorization: SUPERSECRET" | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "hello": "world"
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl 'http://localhost:8080/' -H "Authorization: definitely-not-the-api-key" | jq
    {
        "status": "client_error",
        "error": "Not Authenticated",
        "data": [],
        "pagination": {},
        "input_errors": {}
    }
    ```

    """
    secret_key = clearskies.configs.String(default="")

    """
    The path in our secret manager where an alternate secret can also be fetched

    The alternate secret is exclusively used to authenticate incoming requests.  This allows for secret
    rotation - Point secret_key to a new secret and alternate_secret_key to the old secret.  Both will then
    be accepted and you can migrate your applications to only send the new secret.  Once they are all updated,
    remove the alternate_secret_key:

    ```python
    from types import SimpleNamespace
    import clearskies

    def fetch_secret(path):
        if path == "/path/to/my/secret":
            return "SUPERSECRET"
        if path == "/path/to/alternate/secret":
            return "ALSOOKAY"
        raise KeyError(f"Attempt to fetch non-existent secret: {path}")

    fake_secret_manager = SimpleNamespace(get=fetch_secret)

    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Callable(
            lambda: {"hello": "world"},
            authentication=clearskies.authentication.SecretBearer(
                secret_key="/path/to/my/secret",
                alternate_secret_key="/path/to/alternate/secret",
            ),
        ),
        bindings={
            "secrets": fake_secret_manager,
        },
    )
    wsgi()
    ```

    And when invoked:

    ```bash
    $ curl 'http://localhost:8080/' -H "Authorization: SUPERSECRET" | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "hello": "world"
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl 'http://localhost:8080/' -H "Authorization: ALSOOKAY" | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "hello": "world"
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl 'http://localhost:8080/' -H "Authorization: NOTTHESECRET" | jq
    {
        "status": "client_error",
        "error": "Not Authenticated",
        "data": [],
        "pagination": {},
        "input_errors": {}
    }
    ```

    """
    alternate_secret_key = clearskies.configs.String(default="")

    """
    The name of the environment variable from which we should fetch our key.
    """
    environment_key = clearskies.configs.String(default="")

    """
    The name of an alternate environment variable from which we should fetch our key.

    This allows for secret rotation by allowing the API to accept a secret from two different
    environment variables: an old value and a new value.  You can then migrate your client applications
    to use the new key and, once they are all migrated, remove the old key from the application
    configuration.  Here's an example:

    ```python
    import os
    import clearskies

    os.environ["MY_AUTH_SECRET"] = "SUPERSECRET"
    os.environ["MY_ALT_SECRET"] = "ALSOOKAY"

    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Callable(
            lambda: {"hello": "world"},
            authentication=clearskies.authentication.SecretBearer(
                environment_key="MY_AUTH_SECRET",
                alternate_environment_key="MY_ALT_SECRET",
            ),
        ),
    )
    wsgi()
    ```

    And when invoked:

    ```bash
    $ curl 'http://localhost:8080/' -H "Authorization: SUPERSECRET" | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "hello": "world"
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl 'http://localhost:8080/' -H "Authorization: ALSOOKAY" | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "hello": "world"
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl 'http://localhost:8080/' -H "Authorization: NOTTHESECRET" | jq
    {
        "status": "client_error",
        "error": "Not Authenticated",
        "data": [],
        "pagination": {},
        "input_errors": {}
    }
    ```

    """
    alternate_environment_key = clearskies.configs.String(default="")

    """
    The expected prefix (if any) that should come before the secret key in the authorization header.

    This applies to both the incoming authentication process and outgoing authentication headers.  Some systems
    require a prefix before the auth token in the HTTP header (e.g. `Authorization: TOKEN [auth key here]`).
    You can provide that prefix to `header_prefix` in order for the endpoint to require a prefix or the api backend
    to provide such a prefix.  Note that the prefix is case-insensitive and it does not assume a space between the
    prefix and the token (so, if you want a space, you must explicitly put it in the prefix).  Here's an example:

    ```python
    import os
    import clearskies

    os.environ["MY_AUTH_SECRET"] = "SUPERSECRET"

    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Callable(
            lambda: {"hello": "world"},
            authentication=clearskies.authentication.SecretBearer(
                environment_key="MY_AUTH_SECRET",
                header_prefix="secret-token ",
            ),
        ),
    )
    wsgi()
    ```

    And then usage:

    ```bash
    $ curl 'http://localhost:8080/' -H "Authorization: SECRET-TOKEN SUPERSECRET" | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "hello": "world"
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl 'http://localhost:8080/' -H "Authorization: SUPERSECRET" | jq
    {
        "status": "client_error",
        "error": "Not Authenticated",
        "data": [],
        "pagination": {},
        "input_errors": {}
    }
    ```
    """
    header_prefix = clearskies.configs.String(default="")

    """
    The length of our header prefix
    """
    header_prefix_length = None

    """
    The name of our security scheme in the auto-generated documentation
    """
    documentation_security_name = clearskies.configs.String(default="ApiKey")

    _secret: str = None  #  type: ignore
    _alternate_secret: str = None  # type: ignore

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        secret_key: str = "",
        alternate_secret_key: str = "",
        environment_key: str = "",
        alternate_environment_key: str = "",
        header_prefix: str = "",
        documentation_security_name: str = "",
    ):
        if not secret_key and not environment_key:
            raise ValueError("Must set either 'secret_key' or 'environment_key' when configuring the SecretBearer")
        self.header_prefix_length = len(header_prefix)
        self.finalize_and_validate_configuration()

    @property
    def secret(self):
        if not self._secret:
            self._secret = (
                self.secrets.get(self.secret_key) if self.secret_key else self.environment.get(self.environment_key)
            )
        return self._secret

    def clear_credential_cache(self):
        if self.secret_key:
            self._secret = None  # type: ignore

    @property
    def alternate_secret(self):
        if not self.alternate_secret_key and not self.alternate_environment_key:
            return ""

        if not self._alternate_secret:
            self._alternate_secret = (
                self.secrets.get(self.alternate_secret_key)
                if self.secret_key
                else self.environment.get(self.alternate_environment_key)
            )
        return self._alternate_secret

    def headers(self, retry_auth=False):
        self._configured_guard()
        if retry_auth:
            self.clear_credential_cache()
        return {"Authorization": f"{self.header_prefix}{self.secret}"}

    def authenticate(self, input_output):
        self._configured_guard()
        auth_header = input_output.request_headers.authorization
        if not auth_header:
            return False
        if auth_header[: self.header_prefix_length].lower() != self.header_prefix.lower():
            # self._logging.debug(
            #     "Authentication failure due to prefix mismatch.  Configured prefix: "
            #     + self._header_prefix.lower()
            #     + ".  Found prefix: "
            #     + auth_header[: self._header_prefix_length].lower()
            # )
            return False
        provided_secret = auth_header[self.header_prefix_length :]
        if self.secret == provided_secret:
            # self._logging.debug("Authentication success")
            return True
        if self.alternate_secret and provided_secret == self._alternate_secret:
            # self._logging.debug("Authentication success with alternate secret")
            return True
        # self._logging.debug("Authentication failure due to secret mismatch")
        return False

    def authorize(self, authorization):
        raise ValueError("SecretBearer does not support authorization")

    def set_headers_for_cors(self, cors):
        cors.add_header("Authorization")

    def _configured_guard(self):
        if not self.secret:
            raise ValueError("Attempted to use SecretBearer authentication class without providing the configuration")

    def documentation_request_parameters(self):
        return []

    def documentation_security_scheme(self):
        return {
            "type": "apiKey",
            "name": "authorization",
            "in": "header",
        }

    def documentation_security_scheme_name(self):
        return self.documentation_security_name



from typing import Any

from clearskies.input_outputs.input_output import InputOutput

from .headers import Headers


class Programmatic(InputOutput):
    _body: str | dict[str, Any] | list[Any] = ""
    _request_method: str = ""
    _request_headers: dict[str, Any] = {}
    url: str = ""

    def __init__(
        self,
        url: str = "",
        request_method: str = "GET",
        body: str | dict[str, Any] | list[Any] = "",
        query_parameters: dict[str, Any] = {},
        request_headers: dict[str, str] = {},
    ):
        self.url = url
        self._request_headers = {**request_headers}
        self._body_loaded_as_json = True
        self._body_as_json = None
        self._request_method = request_method
        if body:
            self._body = body
            if isinstance(body, dict) or isinstance(body, list):
                self._body_as_json = body

        super().__init__()
        self.query_parameters = {**query_parameters}

    def respond(self, response, status_code=200):
        return (status_code, response, self.response_headers)

    def get_script_name(self):
        return self.url

    def get_path_info(self):
        return self.url

    def get_full_path(self):
        return self.url

    def get_request_method(self):
        return self._request_method

    def has_body(self):
        return bool(self._body)

    def get_body(self):
        if not self.has_body():
            return ""

        return self._body

    def context_specifics(self):
        return {}

    def get_client_ip(self):
        return "127.0.0.1"

    def get_query_string(self):
        return ""

    def get_request_headers(self):
        return self._request_headers



from typing import Any, Callable

from clearskies.action import Action
from clearskies.query.condition import Condition
from clearskies.validator import Validator

action = Callable[..., dict[str, Any]] | Action
condition = str | Callable | Condition
join = str | Callable[..., str]
validator = Callable[..., str] | Validator
response = str | bytes | dict[str, Any] | list[Any]



from typing import Any as AnyType

from clearskies.configs import config


class Any(config.Config):
    def __set__(self, instance, value: AnyType):
        instance._set_config(self, value)

    def __get__(self, instance, parent) -> AnyType:
        if not instance:
            return self  # type: ignore
        return instance._get_config(self)



from clearskies.configs import select


class ModelColumn(select.Select):
    def __init__(self, model_column_config_name="", required=False, default=None):
        self.required = required
        self.default = default
        self.model_column_config_name = model_column_config_name
        self.model_class = None

    def __set__(self, instance, value: str):
        if value is None:
            return

        if not isinstance(value, str):
            error_prefix = self._error_prefix(instance)
            raise TypeError(
                f"{error_prefix} attempt to set a value of type '{value.__class__.__name__}' to a string parameter"
            )

        instance._set_config(self, value)
        # unlike select, we won't validate the value currently because we won't be able to
        # do that until the finalize_and_validate_configuration phase

    def get_allowed_columns(self, model_class, column_configs):
        return [name for name in column_configs.keys()]

    def my_description(self):
        return "column"

    def get_model_class(self, instance):
        model_class = self.model_class
        if not model_class and self.model_column_config_name:
            model_class = getattr(instance, self.model_column_config_name)
        return model_class

    def set_model_class(self, model_class):
        self.model_class = model_class

    def finalize_and_validate_configuration(self, instance):
        super().finalize_and_validate_configuration(instance)

        # to check for a valid column we need to know the name of the model class.
        # This can be either provided to us directly or we may be given the name of a configuration
        # from which we fetch the model class
        model_class = self.get_model_class(instance)

        # if we don't have one though, no worries - some classes have to provide it later and
        # can trigger validation then.
        if not model_class:
            return

        allowed_columns = self.get_allowed_columns(model_class, model_class.get_columns())

        value = instance._get_config(self)
        if not value:
            return
        if value not in allowed_columns:
            error_prefix = self._error_prefix(instance)
            my_description = self.my_description()
            raise ValueError(
                f"{error_prefix} attempt to set a value of '{value}' but this is not a {my_description} in the specified model class, '{model_class.__name__}'.  Expected values are: '"
                + "', '".join(allowed_columns)
                + "'"
            )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import requests

import clearskies.configurable
from clearskies.authentication.authorization import Authorization

if TYPE_CHECKING:
    from clearskies.security_headers.cors import Cors


class Authentication(clearskies.configurable.Configurable, requests.auth.AuthBase):
    is_public = True
    can_authorize = False
    has_dynamic_credentials = False

    def clear_credential_cache(self) -> None:
        pass

    def headers(self, retry_auth: bool = False) -> dict[str, str]:
        return {}

    def authenticate(self, input_output) -> bool:
        return True

    def authorize(self, authorization: Authorization):
        raise ValueError("Public endpoints do not support authorization")

    def set_headers_for_cors(self, cors: Cors):
        pass

    def documentation_security_scheme(self) -> dict[str, Any]:
        return {}

    def documentation_security_scheme_name(self) -> str:
        return ""

    def __call__(self, request: requests.models.PreparedRequest) -> requests.models.PreparedRequest:
        request.headers = {**request.headers, **self.headers()}  # type: ignore
        return request



from clearskies.di.inject.by_class import ByClass
from clearskies.di.inject.by_name import ByName
from clearskies.di.inject.di import Di
from clearskies.di.inject.environment import Environment
from clearskies.di.inject.input_output import InputOutput
from clearskies.di.inject.now import Now
from clearskies.di.inject.requests import Requests
from clearskies.di.inject.secrets import Secrets
from clearskies.di.inject.utcnow import Utcnow
from clearskies.di.inject.uuid import Uuid

__all__ = [
    "ByClass",
    "ByName",
    "Di",
    "Environment",
    "InputOutput",
    "Now",
    "Requests",
    "Secrets",
    "Utcnow",
    "Uuid",
]



import re


class Join:
    """
    Parses a join clause.

    Note that this expects a few very specific pattern:

     1. [TYPE] JOIN [right_table_name] ON [left_table_name].[left_column_name]=[right_table_name].[right_column_name]
     2. [TYPE] JOIN [right_table_name] AS [alias] ON [alias].[left_column_name]=[right_table_name].[right_column_name]
     3. [TYPE] JOIN [right_table_name] [alias] ON [alias].[left_column_name]=[right_table_name].[right_column_name]

    NOTE: The allowed join types are ["INNER", "OUTER", "LEFT", "RIGHT"]

    NOTE: backticks are optionally allowed around column and table names.

    Examples:
    ```python
    join = Join("INNER JOIN orders ON users.id=orders.user_id")
    print(f"{join.left_table_name}.{join.left_column_name}")  # prints 'users.id'
    print(f"{join.right_table_name}.{join.right_column_name}")  # prints 'orders.user_id'
    print(join.type)  # prints 'INNER'
    print(join.alias)  # prints ''
    print(join.unaliased_table_name)  # prints 'orders'

    join = Join("JOIN some_long_table_name AS new_table ON old_table.id=new_table.old_id")
    print(f"{join.left_table_name}.{join.left_column_name}")  # prints 'old_table.id'
    print(f"{join.right_table_name}.{join.right_column_name}")  # prints 'new_table.old_id'
    print(join.type)  # prints 'LEFT'
    print(join.alias)  # prints 'new_table'
    print(join.unaliased_table_name)  # prints 'some_long_table_name'
    ```
    """

    """
    The name of the table on the left side of the join
    """
    left_table_name: str = ""

    """
    The name of the column on the left side of the join
    """
    left_column_name: str = ""

    """
    The name of the table on the right side of the join
    """
    right_table_name: str = ""

    """
    The name of the column on the right side of the join
    """
    right_column_name: str = ""

    """
    The type of join (LEFT, RIGHT, INNER, OUTER)
    """
    join_type: str = ""

    """
    An alias for the joined table, if needed
    """
    alias: str = ""

    """
    The actual name of the right table, regardless of alias
    """
    unaliased_table_name: str = ""

    """
    The original join string
    """
    _raw_join: str = ""

    """
    The allowed join types
    """
    _allowed_types = ["INNER", "OUTER", "LEFT", "RIGHT"]

    def __init__(self, join: str):
        self._raw_join = join
        # doing this the simple and stupid way, until that doesn't work.  Yes, it is ugly.
        # Splitting this into two regexps for simplicity: this one does not check for an alias
        matches = re.match(
            "(\\w+\\s+)?join\\s+`?([^\\s`]+)`?\\s+on\\s+`?([^`]+)`?\\.`?([^`]+)`?\\s*=\\s*`?([^`]+)`?\\.`?([^`]+)`?",
            join,
            re.IGNORECASE,
        )
        if matches:
            groups = matches.groups()
            alias = ""
            join_type = groups[0]
            right_table = groups[1]
            first_table = groups[2]
            first_column = groups[3]
            second_table = groups[4]
            second_column = groups[5]
        else:
            matches = re.match(
                "(\\w+\\s+)?join\\s+`?([^\\s`]+)`?\\s+(as\\s+)?(\\S+)\\s+on\\s+`?([^`]+)`?\\.`?([^`]+)`?\\s*=\\s*`?([^`]+)`?\\.`?([^`]+)`?",
                join,
                re.IGNORECASE,
            )
            if not matches:
                raise ValueError(f"Specified join condition, '{join}' does not appear to be a valid join statement")
            groups = matches.groups()
            join_type = groups[0]
            right_table = groups[1]
            alias = groups[3]
            first_table = groups[4]
            first_column = groups[5]
            second_table = groups[6]
            second_column = groups[7]

        # which is the left table and which is the right table?
        match_by = alias if alias else right_table
        if first_table == match_by:
            self.left_table_name = second_table
            self.left_column_name = second_column
            self.right_table_name = first_table
            self.right_column_name = first_column
        elif second_table == match_by:
            self.left_table_name = first_table
            self.left_column_name = first_column
            self.right_table_name = second_table
            self.right_column_name = second_column
        else:
            raise ValueError(
                f"Specified join condition, '{join}' was not understandable because the joined table "
                + "is not referenced in the 'on' clause"
            )

        self.join_type = groups[0].strip().upper() if groups[0] else "INNER"
        self.alias = alias
        self.unaliased_table_name = right_table



class Parameter:
    location = ""
    in_body = False

    def __init__(self, definition, description="", required=False):
        self.definition = definition
        self.description = description
        self.required = required



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Any, Callable, Self, Type, overload

import clearskies.parameters_to_properties
import clearskies.typing
from clearskies import configs
from clearskies.columns.datetime import Datetime

if TYPE_CHECKING:
    from clearskies import Model


class Timestamp(Datetime):
    """
    A timestamp column.

    The difference between this and the datetime column is that this stores the datetime
    as a standard unix timestamp - the number of seconds since the unix epoch.

    Also, this **always** assumes the timezone for the timestamp is UTC

    ```python
    import datetime
    import clearskies


    class Pet(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        last_fed = clearskies.columns.Timestamp()


    def demo_timestamp(utcnow: datetime.datetime, pets: Pet) -> dict[str, str | int]:
        pet = pets.create({
            "name": "Spot",
            "last_fed": utcnow,
        })
        return {
            "last_fed": pet.last_fed.isoformat(),
            "raw_data": pet.get_raw_data()["last_fed"],
        }


    cli = clearskies.contexts.Cli(
        clearskies.endpoints.Callable(
            demo_timestamp,
        ),
        classes=[Pet],
    )
    cli()
    ```

    And when invoked it returns:

    ```json
    {
        "status": "success",
        "error": "",
        "data": {"last_fed": "2025-05-18T19:14:56+00:00", "raw_data": 1747595696},
        "pagination": {},
        "input_errors": {},
    }
    ```

    Note that if you pull the column from the model in the usual way (e.g. `pet.last_fed` you get a timestamp,
    but if you check the raw data straight out of the backend (e.g. `pet.get_raw_data()["last_fed"]`) it's an
    integer.
    """

    # whether or not to include the microseconds in the timestamp
    include_microseconds = configs.Boolean(default=False)
    _descriptor_config_map = None

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        include_microseconds: bool = False,
        default: datetime.datetime | None = None,
        setable: datetime.datetime | Callable[..., datetime.datetime] | None = None,
        is_readable: bool = True,
        is_writeable: bool = True,
        is_searchable: bool = True,
        is_temporary: bool = False,
        validators: clearskies.typing.validator | list[clearskies.typing.validator] = [],
        on_change_pre_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_post_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_save_finished: clearskies.typing.action | list[clearskies.typing.action] = [],
        created_by_source_type: str = "",
        created_by_source_key: str = "",
        created_by_source_strict: bool = True,
    ):
        pass

    def from_backend(self, value) -> datetime.datetime | None:
        mult = 1000 if self.include_microseconds else 1
        if not value:
            date = None
        elif isinstance(value, str):
            if not value.isdigit():
                raise ValueError(
                    f"Invalid data was found in the backend for model {self.model_class.__name__} and column {self.name}: a string value was found that is not a timestamp.  It was '{value}'"
                )
            date = datetime.datetime.fromtimestamp(int(value) / mult, datetime.timezone.utc)
        elif isinstance(value, int) or isinstance(value, float):
            date = datetime.datetime.fromtimestamp(value / mult, datetime.timezone.utc)
        else:
            if not isinstance(value, datetime.datetime):
                raise ValueError(
                    f"Invalid data was found in the backend for model {self.model_class.__name__} and column {self.name}: the value was neither an integer, float, string, or datetime object"
                )
            date = value
        return date.replace(tzinfo=datetime.timezone.utc) if date else None

    def to_backend(self, data: dict[str, Any]) -> dict[str, Any]:
        if not self.name in data or isinstance(data[self.name], int) or data[self.name] == None:
            return data

        value = data[self.name]
        if isinstance(value, str):
            if not value.isdigit():
                raise ValueError(
                    f"Invalid data was sent to the backend for model {self.model_class.__name__} and column {self.name}: a string value was found that is not a timestamp. It was '{value}'"
                )
            value = int(value)
        elif isinstance(value, datetime.datetime):
            value = value.timestamp()
        else:
            raise ValueError(
                f"Invalid data was sent to the backend for model {self.model_class.__name__} and column {self.name}: the value was neither an integer, a string, nor a datetime object"
            )

        return {**data, self.name: int(value)}

    @overload
    def __get__(self, instance: None, cls: type) -> Self:
        pass

    @overload
    def __get__(self, instance: Model, cls: type) -> datetime.datetime:
        pass

    def __get__(self, instance, cls):
        return super().__get__(instance, cls)

    def __set__(self, instance, value: datetime.datetime) -> None:
        instance._next_data[self.name] = value

    def input_error_for_value(self, value: str, operator: str | None = None) -> str:
        if not isinstance(value, int):
            return f"'{self.name}' must be an integer"
        return ""

    def values_match(self, value_1, value_2):
        """Compare two values to see if they are the same."""
        return value_1 == value_2



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import clearskies.parameters_to_properties
import clearskies.typing
from clearskies import configs
from clearskies.columns.string import String

if TYPE_CHECKING:
    from clearskies import Model


class CreatedByAuthorizationData(String):
    """
    This column will automatically take data from the authorization data attached to a request and store it in the model upon creation.

    If authorization data isn't available from the context being executed, then you may end up with an error
    (depending on the context).  This is a good thing if you are trying to consistely provide audit information,
    but may be a problem if your model creation needs to happen more flexibly.  Obviously, a pre-requisite for using this
    class is to have an authentication class attached to your endpoint which populates the authorization data.

    NOTE: columns generally also have the `created_by_source_type` and `created_by_source_key` properties that perform
    this exact same function.  Why do we have those properties and this column?  This column works well if we have
    some simple string values that we want to always pull from the authorization data (email, for instance).  The
    properties work better if you need to pull authorization data but it's not just a string type.  An example might
    be if you wanted to pull the user id out of the authorziation data to populate a `BelongsToId` column.  You wouldn't
    use this column because it can't provide all the functionality related to `BelongsToId`, so instead you would
    use the `BelongsToId` column and set `created_by_source_type` to `authorization_data` and `created_by_source_key` to `user_id`.
    Example usage:

    ```python
    class MyModel(clearskies.Model):
        backend = clearskies.backends.MemoryBackend()
        id_column_name = "id"

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        organization_id = clearskies.columns.CreatedByAuthorizationData("organization_id")


    class MyAuthentication(clearskies.authentication.Authentication):
        def authenticate(self, input_output) -> bool:
            # Authenticate the user!
            #
            # This is where you would normally authenticate the user and provide any data about them to the
            # authorization system.  This might mean validating a JWT and then providing the claims as authorization data,
            # or looking up a session id and providing the session data to the authorization system.  In this case,
            # we're just going to return a fixed id for our organization.
            input_output.authorization_data = {
                "organization_id": "my-super-awesome-organization",
            }
            return True


    cli = clearskies.contexts.Cli(
        clearskies.endpoints.Create(
            MyModel,
            writeable_column_names=["name"],
            readable_column_names=["id", "name", "organization_id"],
            authentication=MyAuthentication(),
        ),
        classes=[MyModel],
    )
    cli()
    ```

    And running this will give you something like:

    ```json
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "49816ea4-0956-461e-abd4-03dbde845ba9",
            "name": "Bob",
            "organization_id": "my-super-awesome-organization",
        },
        "pagination": {},
        "input_errors": {},
    }
    ```
    """

    """
    The key inside the authorization data that should be pulled into the column value when the record is created
    """
    authorization_data_key_name = configs.String(required=True)

    """
    Whether or not to throw an error if the key is not present in the authorization data.
    """
    strict = configs.Boolean(default=True)

    """
    Since this column is always populated automatically, it is never directly writeable.
    """
    is_writeable = configs.Boolean(default=False)
    _descriptor_config_map = None

    _allowed_search_operators = ["=", "in", "is not null", "is null", "like"]

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        authorization_data_key_name: str,
        strict: bool = True,
        is_readable: bool = True,
        is_searchable: bool = True,
        is_temporary: bool = False,
        on_change_pre_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_post_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_save_finished: clearskies.typing.action | list[clearskies.typing.action] = [],
    ):
        self.created_by_source_key = authorization_data_key_name
        self.created_by_source_type = "authorization_data"



from typing import Any

from clearskies.configs import config


class AnyDict(config.Config):
    def __set__(self, instance, value: dict[str, Any]):
        if not isinstance(value, dict):
            error_prefix = self._error_prefix(instance)
            raise TypeError(
                f"{error_prefix} attempt to set a value of type '{value.__class__.__name__}' to a parameter that requries a dictionary."
            )
        for key in value.keys():
            if not isinstance(key, str):
                error_prefix = self._error_prefix(instance)
                raise TypeError(f"{error_prefix} attempt to set a dictionary with a non-string key.")
        instance._set_config(self, value)

    def __get__(self, instance, parent) -> dict[str, Any]:
        if not instance:
            return self  # type: ignore
        return instance._get_config(self)



from typing import Callable

import clearskies.typing
from clearskies import configs, parameters_to_properties
from clearskies.columns.string import String


class Select(String):
    """
    A string column but, when writeable via an endpoint, only specific values are allowed.

    Note: the allowed values are case sensitive.

    ```python
    import clearskies


    class Order(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        total = clearskies.columns.Float()
        status = clearskies.columns.Select(["Open", "Processing", "Shipped", "Complete"])


    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Create(
            Order,
            writeable_column_names=["total", "status"],
            readable_column_names=["id", "total", "status"],
        ),
    )
    wsgi()
    ```

    And when invoked:

    ```bash
    $ curl http://localhost:8080 -d '{"total": 125, "status": "Open"}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "22f2c950-6519-4d8e-9084-013455449b07",
            "total": 125.0,
            "status": "Open"
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl http://localhost:8080 -d '{"total": 125, "status": "huh"}' | jq
    {
        "status": "input_errors",
        "error": "",
        "data": [],
        "pagination": {},
        "input_errors": {
            "status": "Invalid value for status"
        }
    }
    ```
    """

    """ The allowed values. """
    allowed_values = configs.StringList(required=True)
    _descriptor_config_map = None

    @parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        allowed_values: list[str],
        default: str | None = None,
        setable: str | Callable[..., str] | None = None,
        is_readable: bool = True,
        is_writeable: bool = True,
        is_searchable: bool = True,
        is_temporary: bool = False,
        validators: clearskies.typing.validator | list[clearskies.typing.validator] = [],
        on_change_pre_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_post_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_save_finished: clearskies.typing.action | list[clearskies.typing.action] = [],
        created_by_source_type: str = "",
        created_by_source_key: str = "",
        created_by_source_strict: bool = True,
    ):
        pass

    def input_error_for_value(self, value: str, operator: str | None = None) -> str:
        return f"Invalid value for {self.name}" if value not in self.allowed_values else ""



from typing import Callable

import clearskies.typing
from clearskies import parameters_to_properties
from clearskies.columns.belongs_to_id import BelongsToId


class BelongsToSelf(BelongsToId):
    """
    This is a standard BelongsToId column except it's used in cases where the model relates to itself.

    This exists because a model can't refer to itself inside it's own class definition.  There are
    workarounds, but having this class is usually quicker for the developer.

    The only difference between this and BelongsToId is that you don't have to provide the parent class.

    See also HasManySelf

    ```python
    from typing import Any

    import clearskies


    class Category(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        parent_id = clearskies.columns.BelongsToSelf()
        parent = clearskies.columns.BelongsToModel("parent_id")
        children = clearskies.columns.HasManySelf()


    def test_self_relationship(categories: Category) -> dict[str, Any]:
        root = categories.create({"name": "Root"})
        sub = categories.create({"name": "Sub", "parent": root})
        subsub_1 = categories.create({"name": "Sub Sub 1", "parent": sub})
        subsub_2 = categories.create({"name": "Sub Sub 2", "parent_id": sub.id})

        return {
            "root_from_child": subsub_1.parent.parent.name,
            "subsubs_from_sub": [subsub.name for subsub in sub.children],
        }


    cli = clearskies.contexts.Cli(
        clearskies.endpoints.Callable(test_self_relationship),
        classes=[Category],
    )

    if __name__ == "__main__":
        cli()
    ```

    Which when invoked returns:

    ```json
    {
        "status": "success",
        "error": "",
        "data": {"root_from_child": "Root", "subsubs_from_sub": ["Sub Sub 1", "Sub Sub 2"]},
        "pagination": {},
        "input_errors": {},
    }
    ```
    """

    _descriptor_config_map = None

    @parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        readable_parent_columns: list[str] = [],
        join_type: str | None = None,
        where: clearskies.typing.condition | list[clearskies.typing.condition] = [],
        default: str | None = None,
        setable: str | Callable | None = None,
        is_readable: bool = True,
        is_writeable: bool = True,
        is_searchable: bool = True,
        is_temporary: bool = False,
        validators: clearskies.typing.validator | list[clearskies.typing.validator] = [],
        on_change_pre_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_post_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_save_finished: clearskies.typing.action | list[clearskies.typing.action] = [],
        created_by_source_type: str = "",
        created_by_source_key: str = "",
        created_by_source_strict: bool = True,
    ):
        pass

    def finalize_configuration(self, model_class, name) -> None:
        """
        Finalize and check the configuration.

        This is an external trigger called by the model class when the model class is ready.
        The reason it exists here instead of in the constructor is because some columns are tightly
        connected to the model class, and can't validate configuration until they know what the model is.
        Therefore, we need the model involved, and the only way for a property to know what class it is
        in is if the parent class checks in (which is what happens here).
        """
        self.parent_model_class = model_class
        super().finalize_configuration(model_class, name)



import datetime

from clearskies.validators.timedelta import Timedelta


class InThePastAtLeast(Timedelta):
    def check_timedelta(self, as_date: datetime.datetime, column_name: str) -> str:
        if as_date > self.utcnow - self.timedelta:
            return f"'{column_name}' must be at least {self.delta_human_friendly()} in the past."
        return ""



import clearskies.typing
from clearskies import configs, parameters_to_properties
from clearskies.columns.string import String


class CreatedByRoutingData(String):
    """
    This column will automatically take data from the route path in the request and store it in the model upon creation.

    If routing data isn't available from the context being executed, then you may end up with an error
    (depending on the context).  This is a good thing if you are trying to consistely provide audit information,
    but may be a problem if your model creation needs to happen more flexibly.

    NOTE: columns generally also have the `created_by_source_type` and `created_by_source_key` properties that perform
    this exact same function.  Why do we have those properties and this column?  This column works well if we have
    some simple string values that we want to always pull from the route path.  The properties work better if you need to
    pull route data but it's not just a string type.  An example might be if you wanted to pull the user id out of the
    route data to populate a `BelongsToId` column.  You wouldn't use this column because it can't provide all the functionality
    related to `BelongsToId`, so instead you would use the `BelongsToId` column and set `created_by_source_type` to `routing_data` and
    `created_by_source_key` to `user_id`.  Example usage:

    ```python
    import clearskies


    class MyModel(clearskies.Model):
        backend = clearskies.backends.MemoryBackend()
        id_column_name = "id"

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        organization_id = clearskies.columns.CreatedByRoutingData("organization_id")


    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Create(
            MyModel,
            url="/{organization_id}",
            writeable_column_names=["name"],
            readable_column_names=["id", "name", "organization_id"],
        ),
        classes=[MyModel],
    )
    wsgi()
    ```

    And if you invoked this:

    ```bash
    $ curl 'http://localhost:8080/my-org-id' -d '{"name":"Bob"}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "3643db8c-c9d4-47ee-a747-8922c59d9e7e",
            "name": "Bob",
            "organization_id": "my-org-id"
        },
        "pagination": {},
        "input_errors": {}
    }
    ```
    """

    """
    The name of the route path to populate the column with.
    """
    routing_path_name = configs.String(required=True)

    """
    Whether or not to throw an error if the key is not present in the header data.
    """
    strict = configs.Boolean(default=True)

    """
    Since this column is always populated automatically, it is never directly writeable.
    """
    is_writeable = configs.Boolean(default=False)
    _descriptor_config_map = None

    _allowed_search_operators = ["=", "in", "is not null", "is null", "like"]

    @parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        routing_path_name: str,
        strict: bool = True,
        is_readable: bool = True,
        is_searchable: bool = True,
        is_temporary: bool = False,
        on_change_pre_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_post_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_save_finished: clearskies.typing.action | list[clearskies.typing.action] = [],
    ):
        self.created_by_source_key = routing_path_name
        self.created_by_source_type = "routing_data"



from __future__ import annotations

import inspect
from collections import OrderedDict
from typing import TYPE_CHECKING, Any, Type

import clearskies.configs
import clearskies.exceptions
from clearskies import authentication, autodoc, typing
from clearskies.endpoints.simple_search import SimpleSearch
from clearskies.functional import string
from clearskies.input_outputs import InputOutput

if TYPE_CHECKING:
    from clearskies import SecurityHeader
    from clearskies.model import Model


class AdvancedSearch(SimpleSearch):
    """
    An endpoint that grants the client extensive control over searching and filtering.

    Rather than accepting URL parameters (like the SimpleSearch endpoint), this endpoint accepts a JSON POST
    body.  Search conditions are specified as a list of dictionaries containing `column`, `operator`, and
    `value`.  It also accepts up to two sort directives.  Of course, while this endpoint supports arbitrary
    searching, it won't work if the backend itself doesn't support it.  The following is the list of allowed
    keys in the JSON body:

    | Name  | Type                 | Description                                                                | Example |
    |-------|----------------------|----------------------------------------------------------------------------|---------|
    | sort  | list[dict[str, str]] | A list of sort directives containing `column` and `direction`              | `{"sort": [ {"column": "age", "direction": "desc} ] }` |
    | limit | int                  | The number of records to return                                            | `{"limit": `100`}` |
    | where | list[dict[str, Any]] | A list of conditions containing `column`, `operator`, and `value`          | `{"where": [ {"column": "age", "operator": ">", "value": 10} ] }` |
    | *     | str, int             | Pagination information.  The key name and value type depend on the backend | `{"start": 100}` |

    Here's an example making use of the AdvancedSearch endpoint:

    ```python
    import clearskies


    class Company(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()


    class User(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        username = clearskies.columns.String()
        age = clearskies.columns.Integer()
        company_id = clearskies.columns.BelongsToId(Company, readable_parent_columns=["id", "name"])
        company = clearskies.columns.BelongsToModel("company_id")


    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.AdvancedSearch(
            model_class=User,
            readable_column_names=["id", "name", "username", "age", "company"],
            sortable_column_names=["name", "username", "age", "company.name"],
            searchable_column_names=["id", "name", "username", "age", "company_id", "company.name"],
            default_sort_column_name="name",
        ),
        bindings={
            "memory_backend_default_data": [
                {
                    "model_class": Company,
                    "records": [
                        {"id": "5-5-5-5", "name": "Bob's Widgets"},
                        {"id": "3-3-3-3", "name": "New Venture"},
                        {"id": "7-7-7-7", "name": "Jane's Cool Stuff"},
                    ],
                },
                {
                    "model_class": User,
                    "records": [
                        {
                            "id": "1-2-3-4",
                            "name": "Bob Brown",
                            "username": "bobbrown",
                            "age": 18,
                            "company_id": "5-5-5-5",
                        },
                        {
                            "id": "1-2-3-5",
                            "name": "Jane Doe",
                            "username": "janedoe",
                            "age": 52,
                            "company_id": "7-7-7-7",
                        },
                        {
                            "id": "1-2-3-6",
                            "name": "Greg",
                            "username": "greg",
                            "age": 37,
                            "company_id": "7-7-7-7",
                        },
                        {
                            "id": "1-2-3-7",
                            "name": "Curious George",
                            "username": "curious",
                            "age": 7,
                            "company_id": "3-3-3-3",
                        },
                    ],
                },
            ],
        },
    )
    wsgi()
    ```

    If you invoke the endpoint without any additional data, it will simply list all records:

    ```bash
    $ curl 'http://localhost:8080/' | jq
    {
        "status": "success",
        "error": "",
        "data": [
            {
                "id": "1-2-3-4",
                "name": "Bob Brown",
                "username": "bobbrown",
                "age": 18,
                "company": {
                    "id": "5-5-5-5",
                    "name": "Bob's Widgets"
                }
            },
            {
                "id": "1-2-3-7",
                "name": "Curious George",
                "username": "curious",
                "age": 7,
                "company": {
                    "id": "3-3-3-3",
                    "name": "New Venture"
                }
            },
            {
                "id": "1-2-3-6",
                "name": "Greg",
                "username": "greg",
                "age": 37,
                "company": {
                    "id": "7-7-7-7",
                    "name": "Jane's Cool Stuff"
                }
            },
            {
                "id": "1-2-3-5",
                "name": "Jane Doe",
                "username": "janedoe",
                "age": 52,
                "company": {
                    "id": "7-7-7-7",
                    "name": "Jane's Cool Stuff"
                }
            }
        ],
        "pagination": {
            "number_results": 4,
            "limit": 50,
            "next_page": {}
        },
        "input_errors": {}
    }
    ```

    Of course you can also sort and paginate.  Keep in mind that pagination is backend-dependent:

    ```bash
    $ curl 'http://localhost:8080/' -d '{"sort":[ {"column": "name", "direction": "desc"} ], "limit": 2, "start": 1}' | jq
    {
        "status": "success",
        "error": "",
        "data": [
            {
                "id": "1-2-3-6",
                "name": "Greg",
                "username": "greg",
                "age": 37,
                "company": {
                    "id": "7-7-7-7",
                    "name": "Jane's Cool Stuff"
                }
            },
            {
                "id": "1-2-3-7",
                "name": "Curious George",
                "username": "curious",
                "age": 7,
                "company": {
                    "id": "3-3-3-3",
                    "name": "New Venture"
                }
            }
        ],
        "pagination": {
            "number_results": 4,
            "limit": 2,
            "next_page": {
            "start": 3
            }
        },
        "input_errors": {}
    }

    ```

    Note that sorting on columns in related models is done via the syntax `relationship_column.column_name`.  These
    must be listed as such in the list of sortable/searchable columns, and then you use the same name to sort/search
    by them:

    ```bash
    $ curl 'http://localhost:8080/' -d '{"sort":[ {"column": "company.name", "direction": "desc"}, {"column": "age", "direction": "asc"} ]}' | jq
    {
        "status": "success",
        "error": "",
        "data": [
            {
                "id": "1-2-3-7",
                "name": "Curious George",
                "username": "curious",
                "age": 7,
                "company": {
                    "id": "3-3-3-3",
                    "name": "New Venture"
                }
            },
                {
                "id": "1-2-3-6",
                "name": "Greg",
                "username": "greg",
                "age": 37,
                "company": {
                    "id": "7-7-7-7",
                    "name": "Jane's Cool Stuff"
                }
            },
                {
                "id": "1-2-3-5",
                "name": "Jane Doe",
                "username": "janedoe",
                "age": 52,
                "company": {
                    "id": "7-7-7-7",
                    "name": "Jane's Cool Stuff"
                }
            },
                {
                "id": "1-2-3-4",
                "name": "Bob Brown",
                "username": "bobbrown",
                "age": 18,
                "company": {
                    "id": "5-5-5-5",
                    "name": "Bob's Widgets"
                }
            }
        ],
        "pagination": {
            "number_results": 4,
            "limit": 50,
            "next_page": {}
        },
        "input_errors": {}
    }

    ```

    And finally searching:

    ```bash
    $ curl 'http://localhost:8080/' -d '{"where":[ {"column": "age", "operator": "<=", "value": 37}, {"column": "username", "operator": "in", "value": ["curious", "greg"]} ]}' | jq
    {
        "status": "success",
        "error": "",
        "data": [
            {
                "id": "1-2-3-7",
                "name": "Curious George",
                "username": "curious",
                "age": 7,
                "company": {
                    "id": "3-3-3-3",
                    "name": "New Venture"
                }
            },
            {
                "id": "1-2-3-6",
                "name": "Greg",
                "username": "greg",
                "age": 37,
                "company": {
                    "id": "7-7-7-7",
                    "name": "Jane's Cool Stuff"
                }
            }
        ],
        "pagination": {
            "number_results": 2,
            "limit": 50,
            "next_page": {}
        },
        "input_errors": {}
    }

    ```

    In terms of the allowed search operators, the standard list of operators is:

     * `<=>`
     * `!=`
     * `<=`
     * `>=`
     * `>`
     * `<`
     * `=`
     * `in`
     * `is not null`
     * `is null`
     * `is not`
     * `is`
     * `like`

    Although not all operators are supported by all columns.  You can use `%` with the `LIKE` operator
    to perform a wildcard search.

    """

    @property
    def allowed_request_keys(self) -> list[str]:
        return self.internal_request_keys

    @property
    def internal_request_keys(self) -> list[str]:
        return ["sort", "limit", "where"]

    def check_request_data(
        self, request_data: dict[str, Any], query_parameters: dict[str, Any], pagination_data: dict[str, Any]
    ) -> None:
        if pagination_data:
            error = self.model.validate_pagination_data(pagination_data, self.auto_case_internal_column_name)
            if error:
                raise clearskies.exceptions.ClientError(error)
        if query_parameters:
            raise clearskies.exceptions.ClientError("Query parameters were found but are not supported.")
        for key in request_data.keys():
            if key not in self.allowed_request_keys:
                raise clearskies.exceptions.ClientError(
                    f"Invalid request parameter found in request body: '{key}'.  Expected parameters: "
                    + ", ".join([self.auto_case_internal_column_name(key) for key in self.allowed_request_keys])
                )
        self.validate_limit(request_data, {})
        sort_key_name = self.auto_case_internal_column_name("sort")
        sort = request_data.get(sort_key_name, [])
        if not isinstance(sort, list):
            raise clearskies.exceptions.ClientError(
                f"'{sort_key_name}' property in request body should be a list, but I found a value of type "
                + sort.__class__.__name
            )
        if sort:
            column_key_name = self.auto_case_internal_column_name("column")
            direction_key_name = self.auto_case_internal_column_name("direction")
            for index, sort_entry in enumerate(sort):
                if not isinstance(sort_entry, dict):
                    raise clearskies.exceptions.ClientError(
                        f"'{sort_key_name}' should be a list of dictionaries, but entry #{index + 1} is a value of type '{sort_entry.__class__.__name}', not a dict"
                    )
                for key_name in [column_key_name, direction_key_name]:
                    if not sort_entry.get(key_name):
                        raise clearskies.exceptions.ClientError(
                            f"Each entry in the sort list should contain both '{column_key_name}' and '{direction_key_name}' but entry #{index + 1} is missing '{key_name}'"
                        )
                    if not isinstance(sort_entry[key_name], str):
                        raise clearskies.exceptions.ClientError(
                            f"{key_name}' must be a string, but for entry #{index + 1} it is a value of type "
                            + sort_entry[key_name].__class__.__name__
                        )
                if sort_entry[direction_key_name].lower() not in ["asc", "desc"]:
                    raise clearskies.exceptions.ClientError(
                        f"{direction_key_name}' must be either 'ASC' or 'DESC', but a different value was found for entry #{index + 1}"
                    )
                if self.auto_case_column_name(sort_entry[column_key_name], False) not in self.sortable_column_names:
                    raise clearskies.exceptions.ClientError(
                        f"Invalid sort column for entry #{index + 1}.  Allowed values are: "
                        + ", ".join(
                            [
                                self.auto_case_column_name(column_name, False)
                                for column_name in self.sortable_column_names
                            ]
                        )
                    )
        where_key_name = self.auto_case_internal_column_name("where")
        where = request_data.get(where_key_name, [])
        if not isinstance(where, list):
            raise clearskies.exceptions.ClientError(
                f"'{where_key_name}' property in request body should be a list, but I found a value of type "
                + where.__class__.__name
            )
        if where:
            column_key_name = self.auto_case_internal_column_name("column")
            operator_key_name = self.auto_case_internal_column_name("operator")
            value_key_name = self.auto_case_internal_column_name("value")
            for index, where_entry in enumerate(where):
                if not isinstance(where_entry, dict):
                    raise clearskies.exceptions.ClientError(
                        f"'{where_key_name}' should be a list of dictionaries, but entry #{index + 1} is a value of type '{where_entry.__class__.__name}', not a dict"
                    )
                for key_name in [column_key_name, operator_key_name, value_key_name]:
                    if key_name not in where_entry:
                        raise clearskies.exceptions.ClientError(
                            f"Each entry in the where list should contain '{column_key_name}', '{operator_key_name}', and '{value_key_name}', but entry #{index + 1} is missing '{key_name}'"
                        )
                    if key_name != value_key_name and not isinstance(where_entry[key_name], str):
                        raise clearskies.exceptions.ClientError(
                            f"{key_name}' must be a string, but for entry #{index + 1} it is a value of type "
                            + sort_entry[key_name].__class__.__name__
                        )
                    if where_entry[column_key_name] not in self.searchable_column_names:
                        raise clearskies.exceptions.ClientError(
                            f"Invalid where column for entry #{index + 1}.  Allowed values are: "
                            + ", ".join(
                                [
                                    self.auto_case_column_name(column_name, True)
                                    for column_name in self.searchable_column_names
                                ]
                            )
                        )
                    [relationship_column_name, column_name] = self.unpack_column_name_with_relationship(
                        self.auto_case_column_name(where_entry[column_key_name], False),
                    )
                    operator = where_entry[operator_key_name].lower()
                    value = where_entry[value_key_name]
                    error_allowed_operators = None
                    if relationship_column_name:
                        column = self.columns[relationship_column_name]
                        if not column.is_allowed_search_operator(operator, relationship_reference=column_name):
                            error_allowed_operators = column.allowed_search_operators(
                                relationship_reference=column_name
                            )
                        else:
                            error = column.check_search_value(
                                value if operator != "in" else value[0],
                                where_entry[operator_key_name],
                                relationship_reference=column_name,
                            )
                    else:
                        column = self.columns[column_name]
                        if not column.is_allowed_search_operator(operator):
                            error_allowed_operators = column.allowed_search_operators()
                        else:
                            error = column.check_search_value(
                                value if operator != "in" else value[0], where_entry[operator_key_name]
                            )
                    if error_allowed_operators:
                        raise clearskies.exceptions.ClientError(
                            f"Invalid operator for entry #{index + 1}.  Allowed operators are: "
                            + ", ".join(column.allowed_search_operators(relationship_reference=column_name))
                        )
                    if error:
                        raise clearskies.exceptions.ClientError(f"Invalid search value for entry #{index + 1}: {error}")

    def configure_model_from_request_data(
        self,
        model: Model,
        request_data: dict[str, Any],
        query_parameters: dict[str, Any],
        pagination_data: dict[str, Any],
    ) -> Model:
        if pagination_data:
            model = model.pagination(**pagination_data)
        sort = request_data.get(self.auto_case_internal_column_name("sort"), [])
        if sort:
            column_key_name = self.auto_case_internal_column_name("column")
            direction_key_name = self.auto_case_internal_column_name("direction")
            model = self.add_join(sort[0][column_key_name], model)
            [primary_table_name, primary_column_name] = self.resolve_references_for_query(sort[0][column_key_name])
            primary_direction = sort[0][direction_key_name]

            if len(sort) > 1:
                [secondary_table_name, secondary_column_name] = self.resolve_references_for_query(
                    sort[1][column_key_name]
                )
                secondary_direction = sort[1][direction_key_name]
            else:
                secondary_column_name = ""
                secondary_direction = ""
                secondary_table_name = ""
            model = model.sort_by(
                primary_column_name if primary_column_name else "",
                primary_direction if primary_direction else "",
                primary_table_name=primary_table_name if primary_table_name else "",
                secondary_column_name=secondary_column_name if secondary_column_name else "",
                secondary_direction=secondary_direction if secondary_direction else "",
                secondary_table_name=secondary_table_name if secondary_table_name else "",
            )
        if request_data.get("limit"):
            model = model.limit(request_data["limit"])

        for where in request_data.get(self.auto_case_internal_column_name("where"), []):
            raw_column_name = self.auto_case_column_name(where[self.auto_case_internal_column_name("column")], False)
            [relationship_column_name, column_name] = self.unpack_column_name_with_relationship(raw_column_name)
            operator = where[self.auto_case_internal_column_name("operator")].lower()
            value = where[self.auto_case_internal_column_name("value")]

            model = self.add_join(raw_column_name, model)
            if relationship_column_name:
                model = self.columns[relationship_column_name].add_search(
                    model, value, operator=operator, relationship_reference=column_name
                )
            else:
                model = self.columns[column_name].add_search(model, value, operator=operator)

        return model

    def documentation_url_search_parameters(self) -> list[autodoc.request.Parameter]:
        return []



from typing import Any, Callable

import clearskies.model
import clearskies.query
from clearskies.autodoc.schema import Integer as AutoDocInteger
from clearskies.autodoc.schema import Schema as AutoDocSchema
from clearskies.backends.backend import Backend
from clearskies.di import InjectableProperties, inject


class CursorBackend(Backend, InjectableProperties):
    """
    The cursor backend connects your models to a MySQL or MariaDB database.

    ## Installing Dependencies

    clearskies uses PyMySQL to manage the database connection and make queries.  This is not installed by default,
    but is a named extra that you can install when needed via:

    ```bash
    pip install clear-skies[mysql]
    ```

    ## Connecting to your server

    By default, database credentials are expected in environment variables:

    | Name        | Default | Value                                                         |
    |-------------|---------|---------------------------------------------------------------|
    | db_host     |         | The hostname where the database can be found                  |
    | db_username |         | The username to connect as                                    |
    | db_password |         | The password to connect with                                  |
    | db_database |         | The name of the database to use                               |
    | db_port     | 3306    | The network port to connect to                                |
    | db_ssl_ca   |         | Path to a certificate to use: enables SSL over the connection |

    However, you can fully control the credential provisioning process by declaring a dependency named `connection_details` and
    setting it to a dictionary with the above keys, minus the `db_` prefix:

    ```python
    class ConnectionDetails(clearskies.di.AdditionalConfig):
        provide_connection_details(self, secrets):
            return {
                "host": secrets.get("database_host"),
                "username": secrets.get("db_username"),
                "password": secrets.get("db_password"),
                "database": secrets.get("db_database"),
                "port": 3306,
                "ssl_ca": "/path/to/ca",
            }

    wsgi = clearskies.contexts.Wsgi(
        some_application,
        additional_configs=[ConnectionDetails()],
        bindings={
            "secrets": "" # some configuration here to point to your secret manager
        }
    )
    ```

    Similarly, some alternate credential provisioning schemes are built into clearskies.  See the
    clearskies.secrets.additional_configs module for those options.

    ## Connecting models to tables

    The table name for your model comes from calling the `destination_name` class method of the model class.  By
    default, this takes the class name, converts it to snake case, and then pluralizes it.  So, if you have a model
    class named `UserPreference` then the cursor backend will look for a table called `user_preferences`.  If this
    isn't what you want, then you can simply override `destination_name` to return whatever table you want:

    ```python
    class UserPreference(clearskies.Model):
        @classmethod
        def destination_name(cls):
            return "some_other_table_name"
    ```

    Additionally, the cursor backend accepts an argument called `table_prefix` which, if provided, will be prefixed
    to your table name.  Finally, you can declare a dependency called `global_table_prefix` which will automatically
    be added to every table name.  In the following example, the table name will be `user_configuration_preferences`
    due to:

     1. The `destination_name` method sets the table name to `preferences`
     2. The `table_prefix` argument to the CursorBackend constructor adds a prefix of `configuration_`
     3. The `global_table_prefix` binding sets a prefix of `user_`, wihch goes before everything else.

    ```python
    import clearskies


    class UserPreference(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.CursorBackend(table_prefix="configuration_")
        id = clearskies.columns.Uuid()

        @classmethod
        def destination_name(cls):
            return "preferences"


    cli = clearskies.contexts.Cli(
        clearskies.endpoints.Callable(
            lambda user_preferences: user_preferences.create(no_data=True).id,
        ),
        classes=[UserPreference],
        bindings={
            "global_table_prefix": "user_",
        },
    )
    ```

    """

    supports_n_plus_one = True
    cursor = inject.ByName("cursor")
    global_table_prefix = inject.ByName("global_table_prefix")
    table_escape_character = "`"
    column_escape_character = "`"
    table_prefix = ""

    def __init__(self, table_escape_character="`", column_escape_character="`", table_prefix=""):
        self.table_escape_character = table_escape_character
        self.column_escape_character = column_escape_character
        self.table_prefix = table_prefix

    def _finalize_table_name(self, table_name):
        table_name = f"{self.global_table_prefix}{self.table_prefix}{table_name}"
        if "." not in table_name:
            return f"{self.table_escape_character}{table_name}{self.table_escape_character}"
        return (
            self.table_escape_character
            + f"{self.table_escape_character}.{self.table_escape_character}".join(table_name.split("."))
            + self.table_escape_character
        )

    def update(self, id: int | str, data: dict[str, Any], model: clearskies.model.Model) -> dict[str, Any]:
        query_parts = []
        parameters = []
        escape = self.column_escape_character
        for key, val in data.items():
            query_parts.append(f"{escape}{key}{escape}=%s")
            parameters.append(val)
        updates = ", ".join(query_parts)

        # update the record
        table_name = self._finalize_table_name(model.destination_name())
        self.cursor.execute(
            f"UPDATE {table_name} SET {updates} WHERE {model.id_column_name}=%s", tuple([*parameters, id])
        )

        # and now query again to fetch the updated record.
        return self.records(
            clearskies.query.Query(
                model.__class__, conditions=[clearskies.query.Condition(f"{model.id_column_name}={id}")]
            )
        )[0]

    def create(self, data: dict[str, Any], model: clearskies.model.Model) -> dict[str, Any]:
        escape = self.column_escape_character
        columns = escape + f"{escape}, {escape}".join(data.keys()) + escape
        placeholders = ", ".join(["%s" for i in range(len(data))])

        table_name = self._finalize_table_name(model.destination_name())
        self.cursor.execute(f"INSERT INTO {table_name} ({columns}) VALUES ({placeholders})", tuple(data.values()))
        new_id = data.get(model.id_column_name)
        if not new_id:
            new_id = self.cursor.lastrowid
        if not new_id:
            raise ValueError("I can't figure out what the id is for a newly created record :(")

        return self.records(
            clearskies.query.Query(
                model.__class__, conditions=[clearskies.query.Condition(f"{model.id_column_name}={new_id}")]
            )
        )[0]

    def delete(self, id: int | str, model: clearskies.model.Model) -> bool:
        table_name = self._finalize_table_name(model.destination_name())
        self.cursor.execute(f"DELETE FROM {table_name} WHERE {model.id_column_name}=%s", (id,))
        return True

    def count(self, query: clearskies.query.Query) -> int:
        (sql, parameters) = self.as_count_sql(query)
        self.cursor.execute(sql, parameters)
        for row in self.cursor:
            return row[0] if type(row) == tuple else row["count"]
        return 0

    def records(
        self, query: clearskies.query.Query, next_page_data: dict[str, str | int] | None = None
    ) -> list[dict[str, Any]]:
        # I was going to get fancy and have this return an iterator, but since I'm going to load up
        # everything into a list anyway, I may as well just return the list, right?
        (sql, parameters) = self.as_sql(query)
        self.cursor.execute(sql, parameters)
        records = [row for row in self.cursor]
        if type(next_page_data) == dict:
            limit = query.limit
            start = query.pagination.get("start", 0)
            if limit and len(records) == limit:
                next_page_data["start"] = int(start) + int(limit)
        return records

    def as_sql(self, query: clearskies.query.Query) -> tuple[str, tuple[Any]]:
        escape = self.column_escape_character
        table_name = query.model_class.destination_name()
        (wheres, parameters) = self.conditions_as_wheres_and_parameters(
            query.conditions, query.model_class.destination_name()
        )
        select_parts = []
        if query.select_all:
            select_parts.append(self._finalize_table_name(table_name) + ".*")
        if query.selects:
            select_parts.extend(query.selects)
        select = ", ".join(select_parts)
        if query.joins:
            joins = " " + " ".join([join._raw_join for join in query.joins])
        else:
            joins = ""
        if query.sorts:
            sort_parts = []
            for sort in query.sorts:
                table_name = sort.table_name
                column_name = sort.column_name
                direction = sort.direction
                prefix = self._finalize_table_name(table_name) + "." if table_name else ""
                sort_parts.append(f"{prefix}{escape}{column_name}{escape} {direction}")
            order_by = " ORDER BY " + ", ".join(sort_parts)
        else:
            order_by = ""
        group_by = self.group_by_clause(query.group_by)
        limit = ""
        if query.limit:
            start = 0
            limit_size = int(query.limit)
            if "start" in query.pagination:
                start = int(query.pagination["start"])
            limit = f" LIMIT {start}, {limit_size}"

        table_name = self._finalize_table_name(table_name)
        return (
            f"SELECT {select} FROM {table_name}{joins}{wheres}{group_by}{order_by}{limit}".strip(),
            parameters,
        )

    def as_count_sql(self, query: clearskies.query.Query) -> tuple[str, tuple[Any]]:
        escape = self.column_escape_character
        # note that this won't work if we start including a HAVING clause
        (wheres, parameters) = self.conditions_as_wheres_and_parameters(
            query.conditions, query.model_class.destination_name()
        )
        # we also don't currently support parameters in the join clause - I'll probably need that though
        if query.joins:
            # We can ignore left joins because they don't change the count
            join_sections = filter(lambda join: join.type != "LEFT", query.joins)  # type: ignore
            joins = " " + " ".join([join._raw_join for join in join_sections])
        else:
            joins = ""
        table_name = self._finalize_table_name(query.model_class.destination_name())
        if not query.group_by:
            query_string = f"SELECT COUNT(*) AS count FROM {table_name}{joins}{wheres}"
        else:
            group_by = self.group_by_clause(query.group_by)
            query_string = (
                f"SELECT COUNT(*) AS count FROM (SELECT 1 FROM {table_name}{joins}{wheres}{group_by}) AS count_inner"
            )
        return (query_string, parameters)

    def conditions_as_wheres_and_parameters(
        self, conditions: list[clearskies.query.Condition], default_table_name: str
    ) -> tuple[str, tuple[Any]]:
        if not conditions:
            return ("", ())  # type: ignore

        parameters = []
        where_parts = []
        for condition in conditions:
            parameters.extend(condition.values)
            table = condition.table_name if condition.table_name else self._finalize_table_name(default_table_name)
            column = condition.column_name
            where_parts.append(
                condition._with_placeholders(
                    f"{table}.{column}",
                    condition.operator,
                    condition.values,
                    escape=False,
                )
            )
        return (" WHERE " + " AND ".join(where_parts), tuple(parameters))  # type: ignore

    def group_by_clause(self, group_by: str) -> str:
        if not group_by:
            return ""
        escape = self.column_escape_character
        if "." not in group_by:
            return f" GROUP BY {escape}{group_by}{escape}"
        parts = group_by.split(".", 1)
        table = parts[0]
        column = parts[1]
        return f" GROUP BY {escape}{table}{escape}.{escape}{column}{escape}"

    def validate_pagination_data(self, data: dict[str, Any], case_mapping: Callable) -> str:
        extra_keys = set(data.keys()) - set(self.allowed_pagination_keys())
        if len(extra_keys):
            key_name = case_mapping("start")
            return "Invalid pagination key(s): '" + "','".join(extra_keys) + f"'.  Only '{key_name}' is allowed"
        if "start" not in data:
            key_name = case_mapping("start")
            return f"You must specify '{key_name}' when setting pagination"
        start = data["start"]
        try:
            start = int(start)
        except:
            key_name = case_mapping("start")
            return f"Invalid pagination data: '{key_name}' must be a number"
        return ""

    def allowed_pagination_keys(self) -> list[str]:
        return ["start"]

    def documentation_pagination_next_page_response(self, case_mapping: Callable[[str], str]) -> list[Any]:
        return [AutoDocInteger(case_mapping("start"), example=0)]

    def documentation_pagination_next_page_example(self, case_mapping: Callable[[str], str]) -> dict[str, Any]:
        return {case_mapping("start"): 0}

    def documentation_pagination_parameters(
        self, case_mapping: Callable[[str], str]
    ) -> list[tuple[AutoDocSchema, str]]:
        return [
            (
                AutoDocInteger(case_mapping("start"), example=0),
                "The zero-indexed record number to start listing results from",
            )
        ]



# type: ignore
from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from clearskies.input_output import InputOutput

from clearskies import exceptions
from clearskies.functional import string


class End(ABC):
    """
    DRY for endpoint and endpoint groups.

    This class is just here to hold some common functionality between Endpoints and EndpointGroups.
    The two classes have plenty of overlap but are different enough that I don't want either to inherit
    from the other.
    """

    def add_url_prefix(self, prefix: str) -> None:
        self.url = (prefix.rstrip("/") + "/" + self.url.lstrip("/")).lstrip("/")

    def top_level_authentication_and_authorization(self, input_output: InputOutput) -> None:
        """
        Handle authentication and authorization for this endpoint.

        In the event of an AuthN/AuthZ issue, raise an exception.  Otherwise, return None
        """
        if not self.authentication:
            return
        try:
            if not self.authentication.authenticate(input_output):
                raise exceptions.Authentication("Not Authenticated")
        except exceptions.ClientError as client_error:
            raise exceptions.Authentication(str(client_error))
        if self.authorization:
            try:
                if not self.authorization.gate(input_output.authorization_data, input_output):
                    raise exceptions.Authorization("Not Authorized")
            except exceptions.ClientError as client_error:
                raise exception.Authorization(str(client_error))

    def __call__(self, input_output: InputOutput) -> Any:
        """
        Execute the endpoint.

        This function mostly just checks AuthN/AuthZ and then passes along control to the handle method.
        It also checks for all the appropriate exceptions from clearskies.exceptions and turns those into the
        expected response.  As a result, when building a new endpoint, you normally modify the handle method
        rather than this one.
        """
        # these two configs can have arbitrary classes attached, which may use injectable properties.  Because they are
        # hiding in configs, the system for automatically discovering these won't work, so we have to manually check them.
        # We can't do this in the constructor because self.di hasn't been populated yet, and we can't do this in
        # our own injectable_properties class method because we need to operate at the instance level
        for config_name in ["authentication", "authorization"]:
            config = getattr(self, config_name)
            if config and hasattr(config, "injectable_properties"):
                config.injectable_properties(self.di)

        response = self.populate_routing_data(input_output)
        if response:
            return response

        self.di.add_binding("input_output", input_output)

        # catch everything when we do an AuthN/AuthZ check because we allow custom-defined classes,
        # and this gives more flexibility (or possibly forgiveness) for how they raise exceptions.
        try:
            self.top_level_authentication_and_authorization(input_output)
        except exceptions.Authentication as auth_error:
            return self.error(input_output, str(auth_error), 401)
        except exceptions.Authorization as auth_error:
            return self.error(input_output, str(auth_error), 403)
        except exceptions.NotFound as not_found:
            return self.error(input_output, str(not_found), 404)
        except exceptions.MovedPermanently as redirect:
            return self.redirect(input_output, str(redirect), 302)
        except exceptions.MovedTemporarily as redirect:
            return self.redirect(input_output, str(redirect), 307)

        try:
            response = self.handle(input_output)
        except exceptions.ClientError as client_error:
            return self.error(input_output, str(client_error), 400)
        except exceptions.InputErrors as input_errors:
            return self.input_errors(input_output, input_errors.errors)
        except exceptions.Authentication as auth_error:
            return self.error(input_output, str(auth_error), 401)
        except exceptions.Authorization as auth_error:
            return self.error(input_output, str(auth_error), 403)
        except exceptions.NotFound as auth_error:
            return self.error(input_output, str(auth_error), 404)
        except exceptions.MovedPermanently as redirect:
            return self.redirect(input_output, str(redirect), 302)
        except exceptions.MovedTemporarily as redirect:
            return self.redirect(input_output, str(redirect), 307)

        return response

    def populate_routing_data(self, input_output: InputOutput) -> Any:
        raise NotImplementedError()

    def add_response_headers(self, input_output: InputOutput) -> None:
        if self.response_headers:
            if callable(self.response_headers):
                response_headers = self.di.call_function(
                    self.response_headers, **input_output.get_context_for_callables()
                )
            else:
                response_headers = self.response_headers

            for index, response_header in enumerate(response_headers):
                if not isinstance(response_header, str):
                    raise TypeError(
                        f"Invalid response header in entry #{index + 1}: the header should be a string, but I was given a type of '{header.__class__.__name__}' instead."
                    )
                parts = response_header.split(":", 1)
                if len(parts) != 2:
                    raise ValueError(
                        f"Invalid response header in entry #{index + 1}: the header should be a string in the form of 'key: value' but the given header did not have a colon to separate key and value."
                    )
                input_output.response_headers.add(parts[0], parts[1])
        for security_header in self.security_headers:
            security_header.set_headers_for_input_output(input_output)

    def respond(self, input_output: InputOutput, response: clearskies.typing.response, status_code: int) -> Any:
        self.add_response_headers(input_output)
        return input_output.respond(response, status_code)

    def respond_json(self, input_output: InputOutput, response_data: dict[str, Any], status_code: int) -> Any:
        if "content-type" not in input_output.response_headers:
            input_output.response_headers.add("content-type", "application/json")
        return self.respond(input_output, self.normalize_response(response_data), status_code)

    def normalize_response(self, response_data: dict[str, Any]) -> dict[str, Any]:
        if "status" not in response_data:
            raise ValueError("Huh, status got left out somehow")
        return {
            self.auto_case_internal_column_name("status"): self.auto_case_internal_column_name(response_data["status"]),
            self.auto_case_internal_column_name("error"): response_data.get("error", ""),
            self.auto_case_internal_column_name("data"): response_data.get("data", []),
            self.auto_case_internal_column_name("pagination"): self.normalize_pagination(
                response_data.get("pagination", {})
            ),
            self.auto_case_internal_column_name("input_errors"): response_data.get("input_errors", {}),
        }

    def normalize_pagination(self, pagination: dict[str, Any]) -> dict[str, Any]:
        # pagination isn't always relevant so if it is completely empty then leave it that way
        if not pagination:
            return pagination
        return {
            self.auto_case_internal_column_name("number_results"): pagination.get("number_results", 0),
            self.auto_case_internal_column_name("limit"): pagination.get("limit", 0),
            self.auto_case_internal_column_name("next_page"): {
                self.auto_case_internal_column_name(key): value
                for (key, value) in pagination.get("next_page", {}).items()
            },
        }

    def auto_case_internal_column_name(self, column_name: str) -> str:
        if self.external_casing:
            return string.swap_casing(column_name, "snake_case", self.external_casing)
        return column_name

    def auto_case_column_name(self, column_name: str, internal_to_external: bool) -> str:
        if not self.internal_casing:
            return column_name
        if internal_to_external:
            return string.swap_casing(
                column_name,
                self.internal_casing,
                self.external_casing,
            )
        return string.swap_casing(
            column_name,
            self.external_casing,
            self.internal_casing,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Self, overload

import clearskies.parameters_to_properties
from clearskies import configs
from clearskies.column import Column
from clearskies.columns import CategoryTree

if TYPE_CHECKING:
    from clearskies import Model


class CategoryTreeChildren(Column):
    """
    Return the child categories from a category tree column.

    See the CategoryTree column for usage examples.

    The ancestors are all direct descendants of a given category.  So, given the following tree:

    ```
    Root/
    ├─ Sub/
    │  ├─ Sub Sub/
    │  │  ├─ Sub Sub Sub/
    ├─ Another Child/

    The children of `Root` are `["Sub", "Another Child"]`.  The children of `Sub Sub` are `["Sub Sub Sub"]`.
    """

    """ The name of the category tree column we are connected to. """
    category_tree_column_name = configs.ModelColumn(required=True)

    is_writeable = configs.Boolean(default=False)
    is_searchable = configs.Boolean(default=False)
    _descriptor_config_map = None

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        category_tree_column_name: str,
    ):
        pass

    def finalize_configuration(self, model_class: type, name: str) -> None:
        """Finalize and check the configuration."""
        getattr(self.__class__, "category_tree_column_name").set_model_class(model_class)
        self.model_class = model_class
        self.name = name
        self.finalize_and_validate_configuration()

        # double check that we are pointed to a category tree column
        category_tree_column = getattr(model_class, self.category_tree_column_name)
        if not isinstance(category_tree_column, CategoryTree):
            raise ValueError(
                f"Error with configuration for {model_class.__name__}.{name}, which is a {self.__class__.__name__}.  It needs to point to a category tree column, and it was told to use {model_class.__name__}.{self.category_tree_column_name}, but this is not a CategoryTree column."
            )

    @overload
    def __get__(self, instance: None, cls: type[Model]) -> Self:
        pass

    @overload
    def __get__(self, instance: Model, cls: type[Model]) -> Model:
        pass

    def __get__(self, model, cls):
        if model is None:
            self.model_class = cls
            return self  # type:  ignore

        # this makes sure we're initialized
        if "name" not in self._config:  # type: ignore
            model.get_columns()

        return self.relatives(model)

    def __set__(self, model: Model, value: Model) -> None:
        raise ValueError(
            f"Attempt to set a value to '{model.__class__.__name__}.{self.name}, but this column is not writeable"
        )

    def relatives(self, model: Model, include_all: bool = False, find_parents: bool = False) -> Model | list[Model]:
        id_column_name = model.id_column_name
        model_id = getattr(model, id_column_name)
        model_table_name = model.destination_name()
        category_tree_column = getattr(self.model_class, self.category_tree_column_name)
        tree_table_name = category_tree_column.tree_model_class.destination_name()
        parent_id_column_name = category_tree_column.tree_parent_id_column_name
        child_id_column_name = category_tree_column.tree_child_id_column_name
        is_parent_column_name = category_tree_column.tree_is_parent_column_name
        level_column_name = category_tree_column.tree_level_column_name

        if find_parents:
            join_on = parent_id_column_name
            search_on = child_id_column_name
        else:
            join_on = child_id_column_name
            search_on = parent_id_column_name

        # if we can join then use a join.
        if category_tree_column.load_relatives_strategy:
            relatives = category_tree_column.parent_model.join(
                f"JOIN {tree_table_name} as tree on tree.{join_on}={model_table_name}.{id_column_name}"
            )
            relatives = relatives.where(f"tree.{search_on}={model_id}")
            if not include_all:
                relatives = relatives.where(f"tree.{is_parent_column_name}=1")
            if find_parents:
                relatives = relatives.sort_by(level_column_name, "asc", "tree")
            return relatives

        # joins only work for SQL-like backends.  Otherwise, we have to pull out our list of ids
        branches = category_tree_column.tree_model.where(f"{search_on}={model_id}")
        if not include_all:
            branches = branches.where(f"{is_parent_column_name}=1")
        if find_parents:
            branches = branches.sort_by(level_column_name, "asc")
        ids = [str(branch.get(join_on)) for branch in branches]

        # Can we search with a WHERE IN() clause?  If the backend supports it, it is probably faster
        if category_tree_column.load_relatives_strategy == "where_in":
            return category_tree_column.parent_model.where(f"{id_column_name} IN ('" + "','".join(ids) + "')")

        # otherwise we have to load each model individually which is SLOW....
        return [category_tree_column.parent_model.find(f"{id_column_name}={id}") for id in ids]



from __future__ import annotations

import datetime
from collections import OrderedDict
from typing import TYPE_CHECKING, Any

import dateparser

import clearskies.configs
import clearskies.di
import clearskies.parameters_to_properties
from clearskies.validator import Validator

if TYPE_CHECKING:
    import clearskies.model


class Timedelta(Validator, clearskies.di.InjectableProperties):
    timedelta = clearskies.configs.Timedelta(default=None)

    utcnow = clearskies.di.inject.Utcnow()

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(self, timedelta: datetime.timedelta):
        self.finalize_and_validate_configuration()

    def delta_human_friendly(self):
        remainder = int(self.timedelta.total_seconds())
        parts = []
        conversion = OrderedDict(
            [
                ("year", 31536000),
                ("day", 86400),
                ("hour", 3600),
                ("minute", 60),
                ("second", 1),
            ]
        )
        for name, num_seconds in conversion.items():
            if num_seconds > remainder:
                continue
            amount = int(remainder / num_seconds)
            remainder -= amount * num_seconds
            parts.append(f"{amount} {name}" + ("s" if amount != 1 else ""))
        return ", ".join(parts)

    def check(self, model: clearskies.model.Model, column_name: str, data: dict[str, Any]) -> str:
        if not data.get(column_name):
            return ""

        as_date = dateparser.parse(data[column_name]) if isinstance(data[column_name], str) else data[column_name]
        if not as_date:
            return f"'{column_name}' was not a valid date"
        if as_date.tzinfo == None:
            as_date = as_date.replace(tzinfo=datetime.timezone.utc)
        return self.check_timedelta(as_date, column_name)

    def check_timedelta(self, as_date: datetime.datetime, column_name: str) -> str:
        return ""



from typing import Any


class Request:
    description: str = ""
    relative_path: str = ""
    request_methods: list[str] = []
    parameters: list[Any] = []
    responses = None
    root_properties = None

    def __init__(
        self, description, responses, relative_path="", request_methods="GET", parameters=None, root_properties=None
    ):
        self.description = description
        self.responses = responses
        self.relative_path = relative_path.lstrip("/")
        self.request_methods = [request_methods] if type(request_methods) == str else request_methods
        self.set_parameters(parameters)
        self.root_properties = root_properties if root_properties is not None else {}

    def set_request_methods(self, request_methods):
        self.request_methods = [request_methods] if type(request_methods) == str else request_methods
        return self

    def prepend_relative_path(self, path):
        self.relative_path = path.rstrip("/") + "/" + self.relative_path.lstrip("/")
        return self

    def append_relative_path(self, path):
        self.relative_path = self.relative_path.rstrip("/") + "/" + path.lstrip("/")
        return self

    def set_parameters(self, parameters=None):
        self.parameters = parameters if parameters else []

    def add_parameter(self, parameter):
        self.parameters.append(parameter)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from clearskies import configurable

if TYPE_CHECKING:
    import clearskies.column
    import clearskies.model


class Validator(ABC, configurable.Configurable):
    is_unique = False
    is_required = False

    def __call__(self, model: clearskies.model.Model, column_name: str, data: dict[str, Any]) -> str:
        return self.check(model, column_name, data)

    @abstractmethod
    def check(self, model: clearskies.model.Model, column_name: str, data: dict[str, Any]) -> str:
        pass

    def additional_write_columns(self, is_create=False) -> dict[str, clearskies.column.Column]:
        return {}



from __future__ import annotations

import inspect
from collections import OrderedDict
from typing import TYPE_CHECKING, Any
from typing import Callable as CallableType

import clearskies.configs
import clearskies.exceptions
from clearskies import authentication, autodoc, typing
from clearskies.endpoint import Endpoint
from clearskies.functional import string
from clearskies.input_outputs import InputOutput

if TYPE_CHECKING:
    from clearskies import Column, SecurityHeader
    from clearskies.authentication import Authentication, Authorization
    from clearskies.model import Model
    from clearskies.schema import Schema


class Callable(Endpoint):
    """
    An endpoint that executes a user-defined function.

    The Callable endpoint does exactly that - you provide a function that will be called when the endpoin is invoked.  Like
    all callables invoked by clearskies, you can request any defined depenndency that can be provided by the clearskies
    framework.

    Whatever you return will be returned to the client.  By default, the return value is sent along in the `data` parameter
    of the standard clearskies response.  To suppress this behavior, set `return_standard_response` to `False`.  You can also
    return an model instance, a model query, or a list of model instances and the callable endpoint will automatically return
    the columns specified in `readable_column_names` to the client.

    Here's a basic working example:

    ```python
    import clearskies


    class User(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()
        id = clearskies.columns.Uuid()
        first_name = clearskies.columns.String()
        last_name = clearskies.columns.String()
        age = clearskies.columns.Integer()


    def my_users_callable(users: User):
        bob = users.create({"first_name": "Bob", "last_name": "Brown", "age": 10})
        jane = users.create({"first_name": "Jane", "last_name": "Brown", "age": 10})
        alice = users.create({"first_name": "Alice", "last_name": "Green", "age": 10})

        return jane


    my_users = clearskies.endpoints.Callable(
        my_users_callable,
        model_class=User,
        readable_column_names=["id", "first_name", "last_name"],
    )

    wsgi = clearskies.contexts.WsgiRef(
        my_users,
        classes=[User],
    )
    wsgi()
    ```

    If you run the above script and invoke the server:

    ```bash
    $ curl 'http://localhost:8080' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "4a35a616-3d57-456f-8306-7c610a5e80e1",
            "first_name": "Jane",
            "last_name": "Brown"
        },
        "pagination": {},
        "input_errors": {}
    }
    ```

    The above example demonstrates returning a model and using readable_column_names to decide what is actually sent to the client
    (note that age is left out of the response).  The advantage of doing it this way is that clearskies can also auto-generate
    OpenAPI documentation using this strategy.  Of course, you can also just return any arbitrary data you want.  If you do return
    custom data, and also want your API to be documented, you can pass a schema along to output_schema so clearskies can document
    it:

    ```python
    import clearskies


    class DogResponse(clearskies.Schema):
        species = (clearskies.columns.String(),)
        nickname = (clearskies.columns.String(),)
        level = (clearskies.columns.Integer(),)


    clearskies.contexts.WsgiRef(
        clearskies.endpoints.Callable(
            lambda: {"species": "dog", "nickname": "Spot", "level": 100},
            output_schema=DogResponse,
        )
    )()
    ```

    """

    """
    The callable to execute when the endpoint is invoked
    """
    to_call = clearskies.configs.Callable(default=None)

    """
    A schema that describes the expected input from the client.

    Note that if this is specified it will take precedence over writeable_column_names and model_class, which
    can also be used to specify the expected input.

    ```python
    import clearskies

    class ExpectedInput(clearskies.Schema):
        first_name = clearskies.columns.String(validators=[clearskies.validators.Required()])
        last_name = clearskies.columns.String()
        age = clearskies.columns.Integer(validators=[clearskies.validators.MinimumValue(0)])

    reflect = clearskies.endpoints.Callable(
        lambda request_data: request_data,
        request_methods=["POST"],
        input_schema=ExpectedInput,
    )

    wsgi = clearskies.contexts.WsgiRef(reflect)
    wsgi()
    ```

    And then valid and invalid requests:

    ```bash
    $ curl http://localhost:8080 -d '{"first_name":"Jane","last_name":"Doe","age":1}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "first_name": "Jane",
            "last_name": "Doe",
            "age": 1
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl http://localhost:8080 -d '{"last_name":10,"age":-1,"check":"cool"}' | jq
    {
        "status": "input_errors",
        "error": "",
        "data": [],
        "pagination": {},
        "input_errors": {
            "age": "'age' must be at least 0.",
            "first_name": "'first_name' is required.",
            "last_name": "value should be a string",
            "check": "Input column check is not an allowed input column."
        }
    }
    ```

    """
    input_schema = clearskies.configs.Schema(default=None)

    """
    Whether or not the return value is meant to be wrapped up in the standard clearskies response schema.

    With the standard response schema, the return value of the function will be placed in the `data` portion of
    the standard clearskies response:

    ```python
    import clearskies

    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Callable(
            lambda: {"hello": "world"},
            return_standard_response=True, # the default value
        )
    )
    wsgi()
    ```

    Results in:

    ```bash
    $ curl http://localhost:8080 | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "hello": "world"
        },
        "pagination": {},
        "input_errors": {}
    }
    ```
    But if you want to build your own response:

    ```python
    import clearskies

    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Callable(
            lambda: {"hello": "world"},
            return_standard_response=False,
        )
    )
    wsgi()
    ```

    Results in:

    ```bash
    $ curl http://localhost:8080 | jq
    {
        "hello": "world"
    }
    ```

    Note that you can also return strings this way instead of objects/JSON.

    """
    return_standard_response = clearskies.configs.Boolean(default=True)

    """
    Set to true if the callable will be returning multiple records (used when building the auto-documentation)
    """
    return_records = clearskies.configs.Boolean(default=False)

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        to_call: CallableType,
        url: str = "",
        request_methods: list[str] = ["GET"],
        model_class: type[clearskies.model.Model] | None = None,
        readable_column_names: list[str] = [],
        writeable_column_names: list[str] = [],
        input_schema: Schema | None = None,
        output_schema: Schema | None = None,
        input_validation_callable: CallableType | None = None,
        return_standard_response: bool = True,
        return_records: bool = False,
        response_headers: list[str | CallableType[..., list[str]]] = [],
        output_map: CallableType[..., dict[str, Any]] | None = None,
        column_overrides: dict[str, Column] = {},
        internal_casing: str = "snake_case",
        external_casing: str = "snake_case",
        security_headers: list[SecurityHeader] = [],
        description: str = "",
        authentication: Authentication = authentication.Public(),
        authorization: Authorization = authentication.Authorization(),
    ):
        # we need to call the parent but don't have to pass along any of our kwargs.  They are all optional in our parent, and our parent class
        # just stores them in parameters, which we have already done.  However, the parent does do some extra initialization stuff that we need,
        # which is why we have to call the parent.
        super().__init__()

        if self.input_schema and not self.writeable_column_names:
            self.writeable_column_names = list(self.input_schema.get_columns().keys())

    def handle(self, input_output: InputOutput):
        if self.writeable_column_names or self.input_schema:
            self.validate_input_against_schema(
                self.get_request_data(input_output),
                input_output,
                self.input_schema if self.input_schema else self.model_class,
            )
        else:
            input_errors = self.find_input_errors_from_callable(input_output.request_data, input_output)
            if input_errors:
                raise clearskies.exceptions.InputErrors(input_errors)
        response = self.di.call_function(self.to_call, **input_output.get_context_for_callables())

        if not self.return_standard_response:
            return input_output.respond(response, 200)

        # did the developer return a model?
        if self.model_class and isinstance(response, self.model_class):
            # and is it a query or a single model?
            if response._data:
                return self.success(input_output, self.model_as_json(response, input_output))
            else:
                # with a query we can also get pagination data, maybe?
                converted_models = [self.model_as_json(model, input_output) for model in response]
                return self.success(
                    input_output,
                    converted_models,
                    number_results=len(response) if response.backend.can_count else None,
                    next_page=response.next_page_data(),
                )

        # or did they return a list of models?
        if isinstance(response, list) and all(isinstance(item, self.model_class) for item in response):
            return self.success(input_output, [self.model_as_json(model, input_output) for model in response])

        # if none of the above, just return the data
        return self.success(input_output, response)

    def documentation(self) -> list[autodoc.request.Request]:
        output_schema = self.output_schema if self.output_schema else self.model_class
        nice_model = string.camel_case_to_words(output_schema.__name__)

        schema_model_name = string.camel_case_to_snake_case(output_schema.__name__)
        output_data_schema = (
            self.documentation_data_schema(output_schema, self.readable_column_names)
            if self.readable_column_names
            else []
        )
        output_autodoc = (
            autodoc.schema.Object(
                self.auto_case_internal_column_name("data"),
                children=output_data_schema,
                model_name=schema_model_name if self.readable_column_names else "",
            ),
        )
        if self.return_records:
            output_autodoc.name = nice_model  # type: ignore
            output_autodoc = autodoc.schema.Array(
                self.auto_case_internal_column_name("data"),
                output_autodoc,
            )  # type: ignore

        authentication = self.authentication
        standard_error_responses = []
        if not getattr(authentication, "is_public", False):
            standard_error_responses.append(self.documentation_access_denied_response())
            if getattr(authentication, "can_authorize", False):
                standard_error_responses.append(self.documentation_unauthorized_response())
        if self.writeable_column_names:
            standard_error_responses.append(self.documentation_input_error_response())

        return [
            autodoc.request.Request(
                self.description,
                [
                    self.documentation_success_response(
                        output_autodoc,  # type: ignore
                        description=self.description,
                        include_pagination=self.return_records,
                    ),
                    *standard_error_responses,
                    self.documentation_generic_error_response(),
                ],
                relative_path=self.url,
                request_methods=self.request_methods,
                parameters=[
                    *self.documentation_request_parameters(),
                    *self.standard_url_parameters(),
                ],
                root_properties={
                    "security": self.documentation_request_security(),
                },
            ),
        ]

    def documentation_request_parameters(self) -> list[autodoc.request.Parameter]:
        if not self.writeable_column_names:
            return []

        return self.standard_json_request_parameters(self.input_schema if self.input_schema else self.model_class)

    def documentation_models(self) -> dict[str, autodoc.schema.Schema]:
        if not self.readable_column_names:
            return {}

        output_schema = self.output_schema if self.output_schema else self.model_class
        schema_model_name = string.camel_case_to_snake_case(output_schema.__name__)

        return {
            schema_model_name: autodoc.schema.Object(
                self.auto_case_internal_column_name("data"),
                children=self.documentation_data_schema(output_schema, self.readable_column_names),
            ),
        }



from .audit import Audit
from .belongs_to_id import BelongsToId
from .belongs_to_model import BelongsToModel
from .belongs_to_self import BelongsToSelf
from .boolean import Boolean
from .category_tree import CategoryTree
from .category_tree_ancestors import CategoryTreeAncestors
from .category_tree_children import CategoryTreeChildren
from .category_tree_descendants import CategoryTreeDescendants
from .created import Created
from .created_by_authorization_data import CreatedByAuthorizationData
from .created_by_header import CreatedByHeader
from .created_by_ip import CreatedByIp
from .created_by_routing_data import CreatedByRoutingData
from .created_by_user_agent import CreatedByUserAgent
from .date import Date
from .datetime import Datetime
from .email import Email
from .float import Float
from .has_many import HasMany
from .has_many_self import HasManySelf
from .has_one import HasOne
from .integer import Integer
from .json import Json
from .many_to_many_ids import ManyToManyIds
from .many_to_many_ids_with_data import ManyToManyIdsWithData
from .many_to_many_models import ManyToManyModels
from .many_to_many_pivots import ManyToManyPivots
from .phone import Phone
from .select import Select
from .string import String
from .timestamp import Timestamp
from .updated import Updated
from .uuid import Uuid

__all__ = [
    "Audit",
    "BelongsToId",
    "BelongsToModel",
    "BelongsToSelf",
    "Boolean",
    "CategoryTree",
    "CategoryTreeAncestors",
    "CategoryTreeChildren",
    "CategoryTreeDescendants",
    "Created",
    "CreatedByAuthorizationData",
    "CreatedByHeader",
    "CreatedByIp",
    "CreatedByRoutingData",
    "CreatedByUserAgent",
    "Date",
    "Datetime",
    "Email",
    "Float",
    "HasMany",
    "HasManySelf",
    "HasOne",
    "Integer",
    "Json",
    "ManyToManyIds",
    "ManyToManyIdsWithData",
    "ManyToManyModels",
    "ManyToManyPivots",
    "Phone",
    "Select",
    "String",
    "Timestamp",
    "Updated",
    "Uuid",
]



class Enum:
    def __init__(self, schema, oai3_schema_resolver):
        self.schema = schema
        self.oai3_schema_resolver = oai3_schema_resolver

    def convert(self):
        return {"nullable": True, "type": "string", "enum": [*self.schema.values, None]}



import clearskies.di.inject as inject
from clearskies.di.additional_config import AdditionalConfig
from clearskies.di.additional_config_auto_import import AdditionalConfigAutoImport
from clearskies.di.di import Di
from clearskies.di.injectable import Injectable
from clearskies.di.injectable_properties import InjectableProperties

__all__ = [
    "AdditionalConfig",
    "AdditionalConfigAutoImport",
    "Di",
    "InjectableProperties",
    "injectInjectable",
]



from clearskies.configs import model_column
from clearskies.functional import string


class ModelToIdColumn(model_column.ModelColumn):
    def __init__(
        self,
        model_column_config_name="",
        source_model_class=None,
        source_model_class_config_name="",
        required=False,
        default=None,
    ):
        self.required = required
        self.default = default
        self.model_column_config_name = model_column_config_name
        self.source_model_class_config_name = source_model_class_config_name
        self.model_class = None
        self.source_model_class = source_model_class

    def finalize_and_validate_configuration(self, instance):
        # we use the model class itself to decide on our value.  However,
        # if we don't have it yet, don't worry.  Someone will eventually tell us
        # what we need to know
        model_class = self.get_model_class(instance)
        if not model_class:
            return
        if self.source_model_class:
            model_class = self.source_model_class
        elif self.source_model_class_config_name:
            model_class = getattr(instance, self.source_model_class_config_name)

        has_config = False
        try:
            if instance._get_config(self):
                has_config = True
        except KeyError:
            pass

        if not has_config:
            instance._set_config(self, string.camel_case_to_snake_case(model_class.__name__) + "_id")

        super().finalize_and_validate_configuration(instance)
